package hep.wired.heprep.graphicspanel;

import hep.graphics.heprep.HepRepAttributeAdapter;
import hep.graphics.heprep.HepRepDefaults;
import hep.graphics.heprep.HepRepInstance;
import hep.graphics.heprep.HepRepIterator;
import hep.graphics.heprep.xml.XMLHepRepReader;
import hep.wired.heprep.services.Attributes;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Stroke;
import java.io.IOException;

import java.util.HashMap;
import java.util.Map;
import org.freehep.graphics2d.VectorGraphics;

/**
 * Cache for subscribed attributes
 * 
 * @author Mark Donszelmann
 * @version $Id: AttributeCache.java 13945 2012-02-27 00:27:14Z onoprien $
 */
public class AttributeCache implements Attributes {

  private static final boolean visibility0;
  private static final boolean pickable0;
  private static final String drawAs0;
  private static final Color color0;
  private static final double lineWidth0;
  private static final double lineWidthMultiplier0 = 1.;
  private static final int lineHasArrow0;
  private static final Color frameColor0;
  private static final double frameWidth0;
  private static final Color fillColor0;
  private static final boolean fill0;
  private static final int markSymbol0;
  private static final double markSize0;
  private boolean visibility;
  private boolean pickable;
  private String drawAs;
  private Color color;
  private double lineWidth;
  private double lineWidthMultiplier;
  private float[] lineStyle;
  private int lineHasArrow;
  private Color frameColor;
  private double frameWidth;
  private Color fillColor;
  private boolean fill;
  private int markSymbol;
  private double markSize;
  
  private Color overrideColor;
  private Color overrideFrameColor;
  private Color overrideFillColor;
  private double overrideLineWidth;
  
  private VectorGraphics graphics;
  private HepRepGraphicsMode mode;
  
  private static final Map<String, float[]> style2stroke;

  static {
    boolean error = false;
    try {
      XMLHepRepReader.readDefaults();
    } catch (IOException e) {
      System.out.println("AttributeCache: could not read heprep defaults from XML file.");
      error = true;
    }

    if (error) {
      visibility0 = true;
      pickable0 = true;
      drawAs0 = "point";
      color0 = Color.WHITE;
      lineWidth0 = 1.0;
      lineHasArrow0 = ARROW_NONE;
      frameColor0 = Color.BLACK;
      frameWidth0 = 1.0;
      fillColor0 = Color.WHITE;
      fill0 = false;
      markSymbol0 = VectorGraphics.SYMBOL_BOX;
      markSize0 = 1.0;
    } else {
      visibility0 = HepRepDefaults.getAttValue("visibility").getBoolean();
      pickable0 = HepRepDefaults.getAttValue("ispickable").getBoolean();
      drawAs0 = HepRepDefaults.getAttValue("drawas").getString();
      color0 = HepRepDefaults.getAttValue("color").getColor();
      lineWidth0 = HepRepDefaults.getAttValue("linewidth").getDouble();
      lineHasArrow0 = getArrowSymbol(HepRepDefaults.getAttValue("linehasarrow").getString().toLowerCase());
      frameColor0 = HepRepDefaults.getAttValue("framecolor").getColor();
      frameWidth0 = HepRepDefaults.getAttValue("framewidth").getDouble();
      fillColor0 = HepRepDefaults.getAttValue("fillcolor").getColor();
      fill0 = HepRepDefaults.getAttValue("fill").getBoolean();
      markSymbol0 = VectorGraphics.getSymbol(HepRepDefaults.getAttValue("markname").getString().toLowerCase());
      markSize0 = HepRepDefaults.getAttValue("marksize").getDouble();
    }
    
    style2stroke = new HashMap<String, float[]>();
    style2stroke.put("solid", null);
    style2stroke.put("dotted", new float[] {3.f, 6.f});
    style2stroke.put("dashed", new float[] {15.f, 15.f});
    style2stroke.put("dotdash", new float[] {15.f, 6.f, 3.f, 6.f});
  }

  /**
   * Creates a cache for subscribed attributes.
   */
  public AttributeCache(VectorGraphics graphics, HepRepGraphicsMode mode,
          HepRepIterator iterator) {
    visibility = visibility0;
    pickable = pickable0;
    drawAs = drawAs0;
    color = color0;
    lineWidth = lineWidth0;
    lineWidthMultiplier = lineWidthMultiplier0;
    lineHasArrow = lineHasArrow0;
    frameColor = frameColor0;
    frameWidth = frameWidth0;
    fillColor = fillColor0;
    fill = fill0;
    markSymbol = markSymbol0;
    markSize = markSize0;

    overrideColor = null;
    overrideFrameColor = null;
    overrideFillColor = null;
    overrideLineWidth = 1.;

    this.graphics = graphics;
    this.mode = mode;
    subscribeTo(iterator);
  }

  public boolean isVisible() {
    return visibility;
  }

  public boolean isPickable() {
    return pickable;
  }

  public String getDrawAs() {
    return drawAs;
  }

  public Color getColor() {
    return overrideColor != null ? overrideColor : color;
  }

  public double getLineWidth() {
    return lineWidth * lineWidthMultiplier * overrideLineWidth;
  }

  public int hasLineArrow() {
    return lineHasArrow;
  }

  public Color getFrameColor() {
    return overrideFrameColor != null ? overrideFrameColor : frameColor;
  }

  public double getFrameWidth() {
    return frameWidth;
  }

  public Color getFillColor() {
    return overrideFillColor != null ? overrideFillColor : fillColor;
  }

  public boolean isFilled() {
    return fill;
  }

  public int getMarkSymbol() {
    return markSymbol;
  }

  public double getMarkSize() {
    return markSize;
  }

  public void overrideColor(Color color) {
    overrideColor = color;
    graphics.setColor(getColor());
  }

  public void overrideFrameColor(Color color) {
    overrideFrameColor = color;
  }

  public void overrideFillColor(Color color) {
    overrideFillColor = color;
  }
  
  public void overrideLineWidth(double lineWidth) {
    overrideLineWidth = lineWidth; 
    graphics.setLineWidth(getLineWidth());
  }

  private void subscribeTo(HepRepIterator iterator) {

    iterator.addHepRepAttributeListener("visibility",
            new HepRepAttributeAdapter() {
              public void setAttribute(HepRepInstance instance, String name, boolean value, int showLabel) {
                visibility = value;
              }
              public void removeAttribute(HepRepInstance instance, String name) {
                visibility = visibility0;
              }
            });

    iterator.addHepRepAttributeListener("isPickable",
            new HepRepAttributeAdapter() {
              public void setAttribute(HepRepInstance instance, String name, boolean value, int showLabel) {
                pickable = value;
              }
              public void removeAttribute(HepRepInstance instance, String name) {
                pickable = pickable0;
              }
            });

    iterator.addHepRepAttributeListener("drawas",
            new HepRepAttributeAdapter() {
              public void setAttribute(HepRepInstance instance, String name, String value, String lowerCaseValue, int showLabel) {
                drawAs = lowerCaseValue;
              }
              public void removeAttribute(HepRepInstance instance, String name) {
                drawAs = drawAs0;
              }
            });

    iterator.addHepRepAttributeListener("color",
            new HepRepAttributeAdapter() {
              public void setAttribute(HepRepInstance instance, String name, Color value, int showLabel) {
                color = value;
                graphics.setColor(color);
              }
              public void removeAttribute(HepRepInstance instance, String name) {
                color = color0;
                graphics.setColor(color);
              }
            });

    iterator.addHepRepAttributeListener("framecolor",
            new HepRepAttributeAdapter() {
              public void setAttribute(HepRepInstance instance, String name, Color value, int showLabel) {
                frameColor = value;
              }
              public void removeAttribute(HepRepInstance instance, String name) {
                frameColor = frameColor0;
              }
            });

    if (mode.useLineWidth) {
      
      iterator.addHepRepAttributeListener("linewidth",
              new HepRepAttributeAdapter() {
                // NOTE needed before freehep-jheprep-2.0.3
//                public void setAttribute(HepRepInstance instance, String name, int value, int showLabel) {
//                  setAttribute(instance, name, (double) value, showLabel);
//                }
                public void setAttribute(HepRepInstance instance, String name, double value, int showLabel) {
                  lineWidth = value * mode.lineWidth;
                  graphics.setLineWidth(getLineWidth());
                }
                public void removeAttribute(HepRepInstance instance, String name) {
                  lineWidth = lineWidth0 * mode.lineWidth;
                  graphics.setLineWidth(getLineWidth());
                }
              });
      
      iterator.addHepRepAttributeListener("linewidthmultiplier",
              new HepRepAttributeAdapter() {
                public void setAttribute(HepRepInstance instance, String name, double value, int showLabel) {
                  lineWidthMultiplier = value;
                  graphics.setLineWidth(getLineWidth());
                }
                public void removeAttribute(HepRepInstance instance, String name) {
                  lineWidthMultiplier = lineWidthMultiplier0;
                  graphics.setLineWidth(getLineWidth());
                }
              });
    
    } else {
      lineWidth = mode.lineWidth;
      graphics.setLineWidth(getLineWidth());
    }
    
    if (mode.useLineStyle) {
      
      iterator.addHepRepAttributeListener("linestyle",
              new HepRepAttributeAdapter() {
                public void setAttribute(HepRepInstance instance, String name, String value, String lowerCaseValue, int showLabel) {
                  lineStyle = style2stroke.get(lowerCaseValue);
                  Stroke s = graphics.getStroke();
                  try {
                    BasicStroke bs = (BasicStroke) s;
                    s = new BasicStroke(bs.getLineWidth(), bs.getEndCap(), bs.getLineJoin(), bs.getMiterLimit(), lineStyle, 0.f);
                    graphics.setStroke(s);
                  } catch (ClassCastException x) {}
                }
                public void removeAttribute(HepRepInstance instance, String name) {
                  lineStyle = null;
                  Stroke s = graphics.getStroke();
                  try {
                    BasicStroke bs = (BasicStroke) s;
                    s = new BasicStroke(bs.getLineWidth(), bs.getEndCap(), bs.getLineJoin(), bs.getMiterLimit());
                    graphics.setStroke(s);
                  } catch (ClassCastException x) {}
                }
              });
      
    }

    iterator.addHepRepAttributeListener("linehasarrow",
            new HepRepAttributeAdapter() {
              public void setAttribute(HepRepInstance instance, String name, String value, String lowerCaseValue, int showLabel) {
                lineHasArrow = getArrowSymbol(lowerCaseValue);
              }
              public void removeAttribute(HepRepInstance instance, String name) {
                lineHasArrow = lineHasArrow0;
              }
            });

    iterator.addHepRepAttributeListener("framewidth",
            new HepRepAttributeAdapter() {
              // NOTE needed before freehep-jheprep-2.0.3
              public void setAttribute(HepRepInstance instance, String name, int value, int showLabel) {
                setAttribute(instance, name, (double) value, showLabel);
              }
              public void setAttribute(HepRepInstance instance, String name, double value, int showLabel) {
                frameWidth = value;
              }
              public void removeAttribute(HepRepInstance instance, String name) {
                frameWidth = frameWidth0;
              }
            });

    iterator.addHepRepAttributeListener("fillcolor",
            new HepRepAttributeAdapter() {
              public void setAttribute(HepRepInstance instance, String name, Color value, int showLabel) {
                fillColor = value;
              }
              public void removeAttribute(HepRepInstance instance, String name) {
                fillColor = fillColor0;
              }
            });

    iterator.addHepRepAttributeListener("fill",
            new HepRepAttributeAdapter() {
              public void setAttribute(HepRepInstance instance, String name, boolean value, int showLabel) {
                fill = mode.fillBoxes ? value : false;
              }
              public void removeAttribute(HepRepInstance instance, String name) {
                fill = fill0;
              }
            });

    iterator.addHepRepAttributeListener("markname",
            new HepRepAttributeAdapter() {
              public void setAttribute(HepRepInstance instance, String name, String value, String lowerCaseValue, int showLabel) {
                markSymbol = VectorGraphics.getSymbol(lowerCaseValue);
              }
              public void removeAttribute(HepRepInstance instance, String name) {
                markSymbol = markSymbol0;
              }
            });

    iterator.addHepRepAttributeListener("marksize",
            new HepRepAttributeAdapter() {
              // NOTE needed before freehep-jheprep-2.0.3
              public void setAttribute(HepRepInstance instance, String name, int value, int showLabel) {
                setAttribute(instance, name, (double) value, showLabel);
              }
              public void setAttribute(HepRepInstance instance, String name, double value, int showLabel) {
                markSize = value;
              }
              public void removeAttribute(HepRepInstance instance, String name) {
                markSize = markSize0;
              }
            });
  }

  private static int getArrowSymbol(String lowerCaseName) {
    if (lowerCaseName.equals("none")) {
      return ARROW_NONE;
    }
    if (lowerCaseName.equals("start")) {
      return ARROW_START;
    }
    if (lowerCaseName.equals("end")) {
      return ARROW_END;
    }
    return ARROW_BOTH;
  }
}
