package hep.wired.heprep.graphicspanel;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.*;
import javax.swing.AbstractAction;
import javax.swing.ButtonGroup;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.KeyStroke;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.MenuEvent;
import javax.swing.event.MenuListener;
import javax.swing.undo.UndoableEditSupport;

import hep.graphics.heprep.HepRep;
import hep.graphics.heprep.HepRepAttValue;
import hep.graphics.heprep.HepRepInstance;
import hep.graphics.heprep.HepRepIterator;
import hep.graphics.heprep.util.HepRepUtil;
import hep.wired.feature.HasBoundingBox;
import hep.wired.feature.Resetable;
import hep.wired.heprep.edit.SetProjection;
import hep.wired.heprep.feature.HasNearestPoint;
import hep.wired.heprep.interaction.CutData;
import hep.wired.heprep.plugin.WiredPlugin;
import hep.wired.heprep.projection.AbstractProjection;
import hep.wired.heprep.projection.ParallelProjection;
import hep.wired.heprep.representation.DrawAsCache;
import hep.wired.heprep.representation.DrawAsPoint;
import hep.wired.heprep.services.DrawAs;
import hep.wired.heprep.services.Projection;
import hep.wired.heprep.util.BoundingBoxGraphics2D;
import hep.wired.heprep.util.Keys;
import hep.wired.heprep.util.NearestPointPathConstructor;
import hep.wired.heprep.util.NullGraphics2D;
import hep.wired.heprep.util.PathGraphics2D;
import hep.wired.plot.WiredPlot;
import hep.wired.services.Feature;
import hep.wired.services.GraphicsPanel;
import hep.wired.services.InteractionHandler;
import hep.wired.services.RecordPlot;
import hep.wired.services.ViewPort;
import hep.wired.util.WiredRegistry;
import hep.wired.viewport.RectangularViewPort;
import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import org.freehep.application.Application;
import org.freehep.application.PropertyUtilities;
import org.freehep.application.studio.Studio;
import org.freehep.graphics2d.BufferedPanel;
import org.freehep.graphics2d.PixelGraphics2D;
import org.freehep.graphics2d.VectorGraphics;
import org.freehep.xml.io.XMLIO;
import org.freehep.xml.io.XMLIOManager;
import org.jdom.Element;

/**
 * Graphics panel to display HepReps.
 * 
 * @author Mark Donszelmann
 * @author Dmitry Onoprienko
 * @version $Id: HepRepPanel.java 13943 2012-02-24 22:31:02Z onoprien $
 */

public class HepRepPanel extends BufferedPanel implements GraphicsPanel,
        Resetable, HasBoundingBox, HasNearestPoint, XMLIO {

 // -- Private parts : ---------------------------------------------------------

  // State

  private HepRepPanelModel _model;
  private ChangeListener _modelListener;

  private HepRep heprep;
  
  // Others

  private HepRepGraphicsMode _mode;
  private Projection projection;
  private ViewPort viewPort;
  private HepRepPanelCommandHandler commandHandler;
  private transient HepRepGraphicsMode usedMode;
  private transient HepRepGraphicsMode fastMode;
  private transient Set /* <String> */ missingDrawAsValues;
  private transient DrawAsCache drawAsCache;
  private transient boolean fast;
  private transient InstanceListener _instanceListener;
  private transient Set/* <HepRepInstance> */ _highlight;
  private static final long nDraws = 10;
  private transient long tMin = 200;
  private transient long tMax = 200;
  private transient long tAverage = 200; // reasonable value
  private transient HepRepAttValue initialPhi, initialTheta;
  private transient HepRepAttValue initialScaleX, initialScaleY, initialScaleZ;
  private transient HepRepAttValue initialTranslateX, initialTranslateY, initialTranslateZ;

  private InteractionHandler _oldHandler;
  private boolean _usingTemporaryHandler;

// -- Constructors : -----------------------------------------------------------

 /** Creates a HepRep panel of dimension 800x600. */
  public HepRepPanel(HepRepGraphicsMode mode) {
    this(mode, null, new Dimension(800, 600), null, null);
  }
  
 /** Creates a HepRep panel of dimension 800x600. */
  public HepRepPanel(HepRepGraphicsMode mode, HepRepPanelModel model) {
    this(mode, null, new Dimension(800, 600), null, model);
  }

  /** Creates a HepRep panel of dimension 800x600 with the specified ViewPort. */
  public HepRepPanel(HepRepGraphicsMode mode, ViewPort viewPort, HepRepPanelModel model) {
    this(mode, viewPort, new Dimension(800, 600), null, model);
  }

  private HepRepPanel() {
    this(null, null, new Dimension(800, 600), null, null);
  }

  private HepRepPanel(HepRepGraphicsMode mode, ViewPort viewPort, Dimension preferredSize, Projection projection, HepRepPanelModel model) {

    super(true);

    this._mode = mode;
    if (viewPort == null) {
      viewPort = new RectangularViewPort(0, 0, 1000, 1000, getWidth(), getHeight());
    }
    this.viewPort = viewPort;
    this.projection = projection;

    Properties user = Application.getApplication().getUserProperties();

    // set default projection if not specified
    if (getProjection() == null) {
      String projectionClassName = PropertyUtilities.getString(user, Keys.heprepProjection, "hep.wired.heprep.projecion.ParallelProjection$XY");
      try {
        Class projectionClass = Class.forName(projectionClassName);
        setProjection((Projection) projectionClass.newInstance());
      } catch (Exception e) {
        setProjection(new ParallelProjection.XY());
      }
    }

    heprep = null;

    setPreferredSize(preferredSize);
    setBackground(_mode.backColor);

    setOpaque(true);

    fastMode = new HepRepGraphicsMode(true);
    fast = false;

    _highlight = null;

    setModel(model == null ? new HepRepPanelModel() : model);

    String actionName = "space pressed";
    getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_SPACE, 0, false),actionName);
    getActionMap().put(actionName, new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        if (_oldHandler == null) _oldHandler = WiredPlugin.getPlugin().getInteractionHandler();
      }
    });
    actionName = "space released";
    getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_SPACE, 0, true),actionName);
    getActionMap().put(actionName, new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        _oldHandler = null;
      }
    });
    for (WiredPlugin.IH handler : WiredPlugin.IH.values()) {
      addIHShortcut(handler);
    }

    actionName = "cp visibility";
    getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_V, 0),actionName);
    getActionMap().put(actionName, new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        WiredPlugin.getPlugin().setControlPanel(WiredPlugin.getPlugin().getInteractionPanel());
      }
    });
    actionName = "cp info";
    getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_I, 0),actionName);
    getActionMap().put(actionName, new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        WiredPlugin.getPlugin().setControlPanel(WiredPlugin.getPlugin().getHepRepInfoPanel());
      }
    });
    actionName = "cp cuts";
    getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_C, 0),actionName);
    getActionMap().put(actionName, new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        WiredPlugin.getPlugin().setControlPanel(WiredPlugin.getPlugin().getCutPanel());
      }
    });

    actionName = "copy view";
    getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_N, InputEvent.CTRL_DOWN_MASK),actionName);
    getActionMap().put(actionName, new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        WiredPlugin.getPlugin().showWired(WiredPlugin.getPlugin().createWiredView(null, heprep, HepRepPanel.this, false), null, true);
      }
    });

    actionName = "clone view";
    getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_J, InputEvent.CTRL_DOWN_MASK),actionName);
    getActionMap().put(actionName, new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        WiredPlugin.getPlugin().showWired(WiredPlugin.getPlugin().createWiredView(null, heprep, HepRepPanel.this, true), null, true);
      }
    });

    actionName = "deselect";
    getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_D, 0),actionName);
    getActionMap().put(actionName, new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        _model.setSelection(Collections.<HepRepInstance>emptyList(), HepRepPanel.this);
      }
    });

    addMouseListener(new MouseAdapter() {
      public void mouseMoved(MouseEvent e) {
        System.out.println("x = "+ e.getX());
      }
    });


  }


// -- Access to state info : ---------------------------------------------------

  public HepRepPanelModel getModel() {
    return _model;
  }

  /** Returns the HepRep instance currently displayed in this panel. */
  public HepRep getRecord() {
    return heprep;
  }


// -- Operations : -------------------------------------------------------------

  public void setModel(HepRepPanelModel model) {
    if (_model != null) {
      _model.removeListener(_modelListener);
    }
    _model = model;
    if (model != null) {
      _modelListener = new ChangeListener() {
        public void stateChanged(ChangeEvent event) {
          if (((HepRepPanelModel.Event)event).contains(HepRepPanelModel.Event.Type.SELECTION)) {
            Set<HepRepInstance> selection = _model.getSelection();
            _highlight = selection.isEmpty() ? null : selection;
          }
          repaint();
        }
      };
      _model.addListener(_modelListener);
      Set<HepRepInstance> selection = _model.getSelection();
      _highlight = selection.isEmpty() ? null : selection;
      repaint();
    }
  }

  public void setHepRepInstanceListener(InstanceListener listener) {
    _instanceListener = listener;
  }

  /** Returns a copy of this panel, attached to this panel's model. */
  public HepRepPanel clone() {
    HepRepGraphicsMode defMode = WiredPlugin.getPlugin().getGraphicsMode();
    HepRepPanel panel = new HepRepPanel(defMode, viewPort.copy(), getPreferredSize(), getProjection().copy(), _model);
    if (_mode != defMode) panel._mode = new HepRepGraphicsMode(_mode, panel);
    panel.tMin = tMin;
    panel.tMax = tMax;
    panel.tAverage = tAverage;
    panel.initialPhi = initialPhi;
    panel.initialTheta = initialTheta;
    panel.initialScaleX = initialScaleX;
    panel.initialScaleY = initialScaleY;
    panel.initialScaleZ = initialScaleZ;
    panel.initialTranslateX = initialTranslateX;
    panel.initialTranslateY = initialTranslateY;
    panel.initialTranslateZ = initialTranslateZ;
    return panel;
  }

  /** Returns a deep copy of this panel. */
  public HepRepPanel copy() {
    HepRepGraphicsMode defMode = WiredPlugin.getPlugin().getGraphicsMode();
    HepRepPanel panel = new HepRepPanel(defMode, viewPort.copy(), getPreferredSize(), getProjection().copy(), new HepRepPanelModel(_model));
    if (_mode != defMode) panel._mode = new HepRepGraphicsMode(_mode, panel);
    return panel;
  }

  /** Sets the HepRep to be displayed in this panel and repaints. */
  public void setRecord(Object object) {

    heprep = (HepRep) object;

    // save reset state if available
    if (heprep != null) {
      HepRepIterator iterator = HepRepUtil.getInstances(heprep.getInstanceTreeList(), null, null, false);
      if (iterator.hasNext()) {
        HepRepInstance instance = iterator.nextInstance();

        initialPhi = instance.getAttValue("ViewPhi");
        initialTheta = instance.getAttValue("ViewTheta");

        initialScaleX = instance.getAttValue("ViewScaleX");
        if (initialScaleX == null) {
          initialScaleX = instance.getAttValue("ViewScale");
        }
        initialScaleY = instance.getAttValue("ViewScaleY");
        initialScaleZ = instance.getAttValue("ViewScaleZ");

        if (initialScaleY == null) {
          initialScaleY = initialScaleX;
        }
        if (initialScaleZ == null) {
          initialScaleZ = initialScaleX;
        }

        initialTranslateX = instance.getAttValue("ViewTranslateX");
        initialTranslateY = instance.getAttValue("ViewTranslateY");
        initialTranslateZ = instance.getAttValue("ViewTranslateZ");
      }
    }

    _model.setRecord(heprep);
//    repaint();
  }

  public Object reset(Object newState) {
    Resetable resetable = (Resetable) projection.getFeature(Resetable.class);
    Object oldState = (resetable != null) ? resetable.reset(newState)
            : null;
    /* FIXME !!!!
    // ViewScale
    if (initialScaleX != null) {
    Scaleable3D scaleable3d = (Scaleable3D) getFeature(Scaleable3D.class);
    if (scaleable3d != null) {
    scaleable3d.setScale(initialScaleX.getDouble(), initialScaleY
    .getDouble(), initialScaleZ.getDouble());
    } else {
    Scaleable2D scaleable2d = (Scaleable2D) getFeature(Scaleable2D.class);
    if (scaleable2d != null) {
    scaleable2d.setScale(initialScaleX.getDouble(),
    initialScaleY.getDouble());
    } else {
    Scaleable scaleable = (Scaleable) getFeature(Scaleable.class);
    if (scaleable != null) {
    scaleable.setScale(initialScaleX.getDouble());
    }
    }
    }
    }

    // ViewRotation
    Rotateable3D rotateable3d = (Rotateable3D) getFeature(Rotateable3D.class);
    if (rotateable3d != null) {
    if (initialPhi != null) {
    rotateable3d.rotate(initialPhi.getDouble(), 0.0, 0.0, 1.0);
    }
    if (initialTheta != null) {
    rotateable3d.rotate(initialTheta.getDouble(), 0.0, 1.0, 0.0);
    }
    }

    // ViewTranslate
    double stx = (initialTranslateX != null) ? initialTranslateX
    .getDouble() : 0;
    double sty = (initialTranslateY != null) ? initialTranslateY
    .getDouble() : 0;
    double stz = (initialTranslateZ != null) ? initialTranslateZ
    .getDouble() : 0;

    Translateable3D translateable3D = (Translateable3D) getFeature(Translateable3D.class);
    if (translateable3D != null) {
    translateable3D.setTranslate(stx, sty, stz);
    } else {
    Translateable2D translateable2D = (Translateable2D) getFeature(Translateable2D.class);
    if (translateable2D != null) {
    translateable2D.setTranslate(stx, sty);
    }
    }
     */

    return oldState;
  }

  /** Sets the projection for this panel. */
  public void setProjection(Projection projection) {
    this.projection = projection;

    // save default
    PropertyUtilities.setString(Application.getApplication().getUserProperties(), Keys.heprepProjection, projection.getClass().getName());

    WiredPlugin plugin = WiredPlugin.getPlugin();

    // enable/disable any buttons due to new projection
//    plugin.getInteractionPanel().setSelected(true);

    // make sure interaction handler is valid
    if ((plugin.getInteractionHandler() != null) && !plugin.getInteractionHandler().isSupportedBy(this)) {
      plugin.setInteractionHandler(WiredPlugin.IH.TRANSLATE.get());
    }

    WiredPlot plot = plugin.getPlot();
    if (plot!=null) plot.setSelected(true);
  }

  public void setBounds(int x, int y, int width, int height) {
    super.setBounds(x, y, width, height);

    if (height < width) {
      viewPort = new RectangularViewPort((width - height) / 2, 0, height,
              height, width, height);
    } else {
      viewPort = new RectangularViewPort(0, (height - width) / 2, width,
              width, width, height);
    }
  }

  public void setFastMode(boolean fast) {
    this.fast = fast;
  }

  public void setSelected(RecordPlot plot, boolean selected) {
//    System.out.println("HepRepPanel.setSelected "+ plot.getName() +"  "+ selected);
    if (selected) {
      if (commandHandler == null) {
        commandHandler = new HepRepPanelCommandHandler(plot);
        UndoableEditSupport undoSupport = plot.getUndoableEditSupport();
        if (undoSupport != null) {
          undoSupport.addUndoableEditListener(commandHandler);
        }
      }
      ((Studio) Application.getApplication()).getCommandTargetManager().add(commandHandler);
      InteractionHandler current = WiredPlugin.getPlugin().getInteractionHandler();
      if (current != plot.getInteractionHandler()) {
        plot.setInteractionHandler(current);
      }
      _oldHandler = null;
      _usingTemporaryHandler = false;
    } else if (commandHandler != null) {
      ((Studio) Application.getApplication()).getCommandTargetManager().remove(commandHandler);
    }
  }

  public JPopupMenu modifyPopupMenu(JPopupMenu menu, final RecordPlot plot) {

    // FIXME WIRED-291: re-add when merging really works based on name.
    // Application.getApplication().getXMLMenuBuilder().mergePopupMenu("wiredPopupMenu",
    // menu);

    menu.addSeparator();

    JMenuItem copyMenuItem = new JMenuItem("New View");
    copyMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_N, ActionEvent.CTRL_MASK));
    copyMenuItem.setToolTipText("Create new view with with initial settings copied from this view");
    copyMenuItem.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        WiredPlugin.getPlugin().showWired(WiredPlugin.getPlugin().createWiredView(null, heprep, HepRepPanel.this, false), null, true);
      }
    });
    menu.add(copyMenuItem);

    JMenuItem cloneMenuItem = new JMenuItem("Clone View");
    cloneMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_J, ActionEvent.CTRL_MASK));
    cloneMenuItem.setToolTipText("Create view that will share settings with this view");
    cloneMenuItem.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        WiredPlugin.getPlugin().showWired(WiredPlugin.getPlugin().createWiredView(null, heprep, HepRepPanel.this, true), null, true);
      }
    });
    menu.add(cloneMenuItem);

    menu.addSeparator();
    
    final JMenuItem graphicsMenuItem = new JMenuItem("Graphics mode...");
    graphicsMenuItem.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        HepRepGraphicsMode defMode = WiredPlugin.getPlugin().getGraphicsMode();
        if (_mode == defMode) {
          _mode = new HepRepGraphicsMode(defMode, HepRepPanel.this);
        }
        final JComponent component = _mode.component();
        JButton apply = new JButton("Apply");
        apply.addActionListener(new ActionListener() {
          public void actionPerformed(ActionEvent e) {
            _mode.apply(component);
          }
        });
        JPanel panel = new JPanel();
        panel.setBorder(BorderFactory.createLoweredBevelBorder());
        panel.add(component);
        Object[] options = new Object[]{"OK", apply, "Cancel"};
        int out = JOptionPane.showOptionDialog(null, panel, "Graphics mode", 
                JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.PLAIN_MESSAGE, null, options, options[0]);
        if (out == JOptionPane.YES_OPTION) _mode.apply(component);
        if (_mode.isSame(defMode)) _mode = defMode;
      }
    });
    menu.add(graphicsMenuItem);

    final JMenu projectionMenu = new JMenu("Projection");
    menu.add(projectionMenu);
    projectionMenu.addMenuListener(new MenuListener() {

      public void menuCanceled(MenuEvent event) {
      }

      public void menuSelected(MenuEvent event) {
        projectionMenu.removeAll();

        ButtonGroup group = new ButtonGroup();

        // FIXME, WIRED-242 pick up from preferences
        String[] names = AbstractProjection.defaultProjections.split(":");

        for (int i = 0; i < names.length; i++) {
          addMenu(group, projectionMenu, names[i]);
        }

        projectionMenu.addSeparator();

        // FIXME, should check if projection is at all available
        Collection allProjections = WiredRegistry.allInstances(Projection.class);
        for (Iterator i = allProjections.iterator(); i.hasNext();) {
          Projection p = (Projection) i.next();
          addMenu(group, projectionMenu, p.getName());
        }
      }

      public void menuDeselected(MenuEvent event) {
        projectionMenu.removeAll();
      }

      //
      // Add a menu/item for the part of the name between '$' (or start of
      // the string) and '.' or end of the string.
      // Recursive!
      //
      private void addMenu(ButtonGroup group, JMenu menu, String name) {
        int dollar = name.lastIndexOf('$');
        int dot = name.indexOf('.', dollar + 1);
        String menuName = dot < 0 ? name.substring(dollar + 1) : name.substring(dollar + 1, dot);
        if (dot >= 0) {
          // add submenu or to submenu
          JMenu subMenu = null;
          for (int i = 0; i < menu.getItemCount(); i++) {
            JMenuItem item = menu.getItem(i);
            if ((item instanceof JMenu)
                    && item.getText().equals(menuName)) {
              subMenu = (JMenu) item;
              break;
            }
          }
          if (subMenu == null) {
            subMenu = new JMenu(menuName);
            menu.add(subMenu);
          }
          addMenu(group, subMenu, name.substring(0, dot) + "$"
                  + name.substring(dot + 1));
        } else {
          // add here
          JMenuItem item = new JRadioButtonMenuItem(menuName);
          menu.add(item);
          group.add(item);
          item.setEnabled(false);

          // lookup in projection services
          final Projection projection = (Projection) WiredRegistry.lookup(Projection.class, name);
          if (projection != null) {
            item.setSelected(name.equals(getProjection().getID()));
            item.setEnabled(true);
            item.addActionListener(new ActionListener() {

              public void actionPerformed(ActionEvent event) {
                plot.postEdit(new SetProjection(projection));
              }
            });
          }
        }
      }
    });

    return menu;
  }

  public void setGraphicsMode(HepRepGraphicsMode mode) {
    _mode = mode;
    if (isVisible()) repaint();
  }

// -- Getters : ----------------------------------------------------------------

  public Set/* <HepRepType> */ getVisibleTypes() {
    return _model.getTypesModel().getMarkedObjects(_model.isTypesDepthLimited() ? _model.getTypesDepth() : Integer.MAX_VALUE);
  }

  public Set/* <HepRepInstance> */ getVisibleInstances() {
    return _model.getInstancesModel().getMarkedObjects(_model.isInstancesDepthLimited() ? _model.getInstancesDepth() : Integer.MAX_VALUE);
  }

  public boolean supports(Class featureClass) {
    return getFeature(featureClass) != null;
  }

  /**
   * Returns the feature for a specific featureClass, or null if that
   * featureClass if not in the projection of this panel.
   */
  public Feature getFeature(Class featureClass) {
    if (featureClass.isAssignableFrom(this.getClass())) {
      return this;
    }
    return projection.getFeature(featureClass);
  }

  /** Returns the projection for this panel. */
  public Projection getProjection() {
    return projection;
  }

  /**
   * Returns the bounding box for this plot in its current
   * projection/viewpoint.
   */
  public Rectangle2D getBoundingBoxForPlot() {
    BoundingBoxGraphics2D bb = new BoundingBoxGraphics2D(true);
    draw(bb, true, false);
    return bb.getBoundingBox();
  }

  public Point2D getNearestPoint(Point2D p) {
    NearestPointPathConstructor pathConstructor = new NearestPointPathConstructor(
            p);
    PathGraphics2D pathGraphics = new PathGraphics2D(pathConstructor);
    draw(pathGraphics, true, false);
    return pathConstructor.getNearestPoint();
  }

  /** Returns the viewport for this panel. */
  public ViewPort getViewPort() {
    return viewPort;
  }

  /** Returns the GraphicsMode. */
  public HepRepGraphicsMode getGraphicsMode() {
    return _mode;
  }

  /**
   * Returns a set of DrawAs values for which no DrawAs service was available.
   * These were drawn as Points.
   */
  public Set/* <String> */ getMissingDrawAsValues() {
    NullGraphics2D ng = new NullGraphics2D();
    draw(ng, true, false);
    return missingDrawAsValues;
  }


// -- Painting and drawing : ---------------------------------------------------

  public void repaint() {
    // setSelected toolbar...
    if (commandHandler != null) commandHandler.setChanged();
    if (_mode != null) setBackground(_mode.backColor);
    super.repaint();
  }

  /** Paints the HepRep into this panel. */
  public void paintComponent(VectorGraphics graphics) {
    try {
      // clear background
      Color c = graphics.getColor();
      graphics.setColor(getBackground());
      Rectangle r = getBounds();
      graphics.fillRect(r.x, r.y, r.width, r.height);
      graphics.setColor(c);

      // FIXME: WIRED-380, blitting does not always work
      graphics.setRenderingHint(PixelGraphics2D.KEY_SYMBOL_BLIT,
              PixelGraphics2D.VALUE_SYMBOL_BLIT_OFF);

      // set miterlimit to 1 to avoid jaggy edges
      BasicStroke cs = (BasicStroke) graphics.getStroke();
      graphics.setStroke(new BasicStroke(cs.getLineWidth(), cs.getEndCap(), cs.getLineJoin(), 1.0f, cs.getDashArray(), cs.getDashPhase()));

      long t = System.currentTimeMillis();

      // now draw all the elements
      draw(graphics, fast, false);

      // do some timing measurements, calculating approximate moving
      // averages, minima and maxima
      if (false) {
        t = System.currentTimeMillis() - t;
        if (t != 0) {
          long tNow = t / nDraws;
          tMin = t < tAverage ? tMin + tNow - tMin / nDraws : tMin;
          tMax = t > tAverage ? tMax + tNow - tMax / nDraws : tMax;
          tAverage = tAverage + tNow - tAverage / nDraws;
          if (usedMode != fastMode) {
            System.err.println("Draw took: " + tMin + " <= "
                    + tAverage + " <= " + tMax
                    + " ms averaged over " + nDraws + " redraws.");
          }
        }
      }
    } catch (Throwable t) {
      // FIXME, we need to disentangle this from Application...
      // Application.getApplication().error(t.toString(), t);
      System.err.println(t.toString());
      t.printStackTrace();
    }
  }

  private void draw(VectorGraphics graphics, boolean fast, boolean pick) {
    if (heprep == null) {
      return;
    }
    List layers = (fast || !_mode.useLayering) ? null : heprep.getLayerOrder();
    draw(graphics, fast, layers, null, pick);
  }

  public void draw(VectorGraphics graphics, boolean fast, List layers, Set types, boolean pick) {

    if (heprep == null) return;

    // make types the intersection of types and selectedTypes.
    if (types == null) {
      types = getVisibleTypes();
    } else {
      types = new HashSet(types);
      types.retainAll(getVisibleTypes());
    }

    Set instances = getVisibleInstances();

    usedMode = (fast) ? fastMode : _mode;
    missingDrawAsValues = new HashSet();

    // create a new Cache
    drawAsCache = new DrawAsCache();

    // set rendering mode
    if (usedMode.antiAlias) {
      graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
              RenderingHints.VALUE_ANTIALIAS_ON);
      graphics.setRenderingHint(RenderingHints.KEY_RENDERING,
              RenderingHints.VALUE_RENDER_QUALITY);
    } else {
      graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
              RenderingHints.VALUE_ANTIALIAS_OFF);
      graphics.setRenderingHint(RenderingHints.KEY_RENDERING,
              RenderingHints.VALUE_RENDER_SPEED);
    }

    draw(graphics, heprep.getInstanceTreeList(), layers, types, instances, pick);

    if (!missingDrawAsValues.isEmpty()) {
      System.err.print("Missing DrawAs values: ");
      for (Iterator i = missingDrawAsValues.iterator(); i.hasNext();) {
        System.err.print(i.next() + " ");
      }
      System.err.println();
    }
  }

  private void draw(VectorGraphics graphics, List/* <HepRepInstanceTree> */ trees, List layers, Set types, Set instances, boolean pick) {

    boolean isModified = false;
    boolean modifySelected = (_highlight != null) && _mode.selectionModify;
    boolean modifyUnselected = (_highlight != null) && !_mode.selectionModify;

    HepRepIterator iterator = HepRepUtil.getInstances(trees, layers, types, usedMode.drawFrames);
    AttributeCache atts = new AttributeCache(graphics, usedMode, iterator);

    CutData cutData = null;
    boolean passCut = true;
    if (_model.getCutSet().isEnabled()) {
      Set<String> enabledConditionNames = _model.getCutSet().getConditionNames();
      if (!enabledConditionNames.isEmpty()) {
        cutData = new CutData();
        cutData.subscribeTo(iterator, enabledConditionNames);
      }
    }

    projection.setViewPort(viewPort);

    int i = 0;
    while (iterator.hasNext()) {
      HepRepInstance instance = iterator.nextInstance();

      if ((instances != null) && !instances.contains(instance)) continue;

      if (cutData != null) {
        if (cutData.hasChanged()) {
          passCut = _model.getCutSet().pass(cutData.getData(), instance.getType());
        }
        if (!passCut) continue;
      }

      if (pick && !atts.isPickable()) {
        continue;
      }
      if (atts.isVisible()) {
        if (_instanceListener != null) {
          _instanceListener.currentHepRepInstance(instance);
        }

        // highlight if necessary, and draw

        if (modifySelected || modifyUnselected) {
          boolean isSelected = _highlight.contains(instance);
          if (modifySelected && isSelected) {
            if (_mode.selectionChangeColorSelected) {
              atts.overrideColor(_mode.selectionColorSelected);
              atts.overrideFrameColor(_mode.selectionColorSelected);
              atts.overrideFillColor(_mode.selectionColorSelected);
            }
            atts.overrideLineWidth(_mode.selectionLineWidthMultiplier);
            isModified = true;
          } else if (modifyUnselected && !isSelected) {
            atts.overrideColor(_mode.selectionColorUnselected);
            atts.overrideFrameColor(_mode.selectionColorUnselected);
            atts.overrideFillColor(_mode.selectionColorUnselected);
            isModified = true;
          } else {
            isModified = false;
          }
        }
        draw(graphics, atts, instance, iterator.drawAsFrame());
        i++;
        if (isModified) {
          atts.overrideColor(null);
          atts.overrideFrameColor(null);
          atts.overrideFillColor(null);
          atts.overrideLineWidth(1.);
        }
      }
    }
  }

  private void draw(VectorGraphics graphics, AttributeCache atts, HepRepInstance instance, boolean drawAsFrame) {
    DrawAs handler = drawAsCache.lookup(atts.getDrawAs());
    if (handler == null) {
      handler = drawAsCache.lookup("point");
      if (handler == null) handler = new DrawAsPoint();
      missingDrawAsValues.add(atts.getDrawAs());
    }
    if (drawAsFrame) {
      handler.frame(graphics, instance, atts, usedMode, projection, viewPort);
    } else {
      handler.draw(graphics, instance, atts, usedMode, projection, viewPort);
    }
  }

  public static interface InstanceListener {
    public void currentHepRepInstance(HepRepInstance instance);
  }


// -- XMLIO : ------------------------------------------------------------------

  public void save(XMLIOManager xmlioManager, org.jdom.Element nodeEl) {
    nodeEl.addContent(xmlioManager.save(_mode));
    nodeEl.addContent(xmlioManager.save(projection));
    nodeEl.addContent(xmlioManager.save(viewPort));
  }

  public void restore(XMLIOManager xmlioManager, org.jdom.Element nodeEl) {
    Iterator i = nodeEl.getChildren().iterator();
    _mode = (HepRepGraphicsMode) xmlioManager.restore((Element) i.next());
    projection = (Projection) xmlioManager.restore((Element) i.next());
    viewPort = (ViewPort) xmlioManager.restore((Element) i.next());
  }

// -- Local methods - keyboard shortcuts : -------------------------------------

  private void ihKeyPressed(InteractionHandler handler) {
    if (_usingTemporaryHandler || _oldHandler == null) return;
    _usingTemporaryHandler = true;
    WiredPlugin.getPlugin().setInteractionHandler(handler);
  }

  private void ihKeyReleased(InteractionHandler handler) {
    if (_usingTemporaryHandler) {
      if (handler == WiredPlugin.getPlugin().getInteractionHandler()) {
        if (_oldHandler != null) {
          WiredPlugin.getPlugin().setInteractionHandler(_oldHandler);
        }
        _usingTemporaryHandler = false;
      }
    } else {
      _oldHandler = null;
      WiredPlugin.getPlugin().setInteractionHandler(handler);
    }
  }

  private void addIHShortcut(final WiredPlugin.IH ih) {
    int keyCode = ih.keyCode();
    String name = ih.name() + "released";
    getInputMap().put(KeyStroke.getKeyStroke(keyCode, 0, true),name);
    getActionMap().put(name, new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        ihKeyReleased(ih.get());
      }
    });
    name = ih.name() + "pressed";
    getInputMap().put(KeyStroke.getKeyStroke(keyCode, 0, false),name);
    getActionMap().put(name, new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        ihKeyPressed(ih.get());
      }
    });
  }

}
