package hep.wired.heprep.graphicspanel;

import java.util.*;

import hep.graphics.heprep.HepRep;
import hep.graphics.heprep.HepRepInstance;
import hep.graphics.heprep.HepRepType;
import hep.wired.cut.CutSet;
import hep.wired.heprep.interaction.Flag;
import hep.wired.heprep.tree.HepRepTreeModel;
import hep.wired.heprep.util.WiredHepRepUtil;
import hep.wired.util.tree.WiredTreeModel;
import hep.wired.util.tree.WiredTreeNode;
import javax.swing.JOptionPane;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;
import org.freehep.application.Application;
import org.freehep.application.PropertyUtilities;

/**
 * Class that encapsulates state and settings associated with {@link HepRepPanel}.
 *
 * @author Dmitry Onoprienko
 * @version $Id: $
 */
public class HepRepPanelModel implements WiredTreeModel.MarkListener {

// -- Private parts : ----------------------------------------------------------

  static private String _keyTypeTree = "hep.wired.typetree";
  static private String _keyPickLayers = "hep.wired.picklayers";
  static private String _keyPickTree = "hep.wired.picktree";
  static private String _keyShowAtt = "hep.wired.showatt";

  private String _name;
  private int _nPanel; // number of HepRepPanel that have been associated with this model

  // Current heprep

  private HepRep _heprep;

  // Selected instances and updater :

  private Set<HepRepInstance> _selection; // selected instances
  private boolean _instanceTreeSelectionListenerEnabled = true;
  private TreeSelectionListener _instanceTreeSelectionListener = new TreeSelectionListener() {
    public void valueChanged(TreeSelectionEvent event) {
      if (_instanceTreeSelectionListenerEnabled) {
        _selection = null;
        fireEvent(_instances, Event.Type.SELECTION);
      }
    }
  };

  // Tree panel data :

  private HepRepTreeModel _types;  // visible types
  private int _typesDepth;  // hide types below this level in the tree, if enabled
  private int _typesMaxDepth; // depth of type tree
  private boolean _typesDepthLimited; // hiding types below a certain level is enabled

  private HepRepTreeModel _instances;  // visible instances
  private int _instancesDepth;  // hide instancess below this level in the tree, if enabled
  private int _instancesMaxDepth; // depth of instance tree
  private boolean _instancesDepthLimited; // hiding instances below a certain level is enabled

  private boolean _applyImmediately; // apply visibility changes immediately

  // HepRepInfoPanel and picking data :
  
  private Set<HepRepInstance> _pick; // picked instances

  private List<Flag> _pickLayers;  // pickable layers
  private HepRepTreeModel _pickTypes; // pickable types

  private List<Flag> _attCategories;  // heprep attribute categories to be displayed

  // Cuts data :

  private CutSet _cutSet;

  // Listeners :

  private IdentityHashMap<ChangeListener, EnumSet<Event.Type>> _listeners;

// -- Constructors and copy : --------------------------------------------------

  public HepRepPanelModel() {
    this("");
  }

  public HepRepPanelModel(String name) {
    _name = name;
    _nPanel = 1;
    _listeners = new IdentityHashMap();
    _cutSet = new CutSet();
    _cutSet.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        fireEvent(Event.Type.CUTS);
      }
    });
    setDefaults();
  }

  public HepRepPanelModel(HepRepPanelModel model) {
    _name = model._name;
    _nPanel = 1;
    _heprep = model._heprep;
    _listeners = new IdentityHashMap();
    _cutSet = new CutSet(model._cutSet);
    _cutSet.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        fireEvent(Event.Type.CUTS);
      }
    });
    _types = new HepRepTreeModel(model._types);
    _typesDepth = model._typesDepth;
    _typesMaxDepth = model._typesMaxDepth;
    _typesDepthLimited = model._typesDepthLimited;
    _instances = new HepRepTreeModel(model._instances);
    _instancesDepth = model._instancesDepth;
    _instancesMaxDepth = model._instancesMaxDepth;
    _instancesDepthLimited = model._instancesDepthLimited;
    _applyImmediately = model._applyImmediately;
    _pick = Collections.emptySet();
//    _pick = model._pick.isEmpty() ? Collections.emptySet() : new HashSet(model._pick);
    _pickLayers = model._pickLayers.isEmpty() ? Collections.emptyList() : new ArrayList(model._pickLayers);
    _pickTypes = new HepRepTreeModel(model._pickTypes);
    _attCategories = model._attCategories.isEmpty() ? Collections.emptyList() : new ArrayList(model._attCategories);
    setSelection(model.getSelection(), null);
  }

  private void setDefaults() {
    _types = new HepRepTreeModel(null, true);
    _typesDepth = -1;
    _typesMaxDepth = 0;
    _typesDepthLimited = false;
    _instances = new HepRepTreeModel(null, false);
    _instancesDepth = -1;
    _instancesMaxDepth = 0;
    _instancesDepthLimited = false;
    _applyImmediately = true;
    _pick = Collections.emptySet();
    _pickLayers = Collections.emptyList();
    _pickTypes = new HepRepTreeModel(null, true);
    _attCategories = Collections.emptyList();
  }

// -- Modifying constituent models and synchronizing changes : -----------------

  public void markChanged(WiredTreeModel.MarkEvent e) {
    if (e.getSource() == _types) {
      Set<? extends Object> visibleTypes = _types.getMarkedObjects();
      syncInstanceTypeTrees(visibleTypes);
      updatePickTree(visibleTypes);
      _instances.fireMarkChanged();
    } else if (e.getSource() == _instances) {
      fireEvent(Event.Type.VISIBILITY);
    }
  }

  public void setTypesDepth(int depth) {
    _typesDepth = depth;
    if (_typesDepthLimited) {
      updateTypeTree();
      _types.fireMarkChanged();
    }
  }

  public void setTypesDepth(boolean limited) {
    _typesDepthLimited = limited;
    if ( _typesDepth < _typesMaxDepth) {
      updateTypeTree();
      _types.fireMarkChanged();
    }
  }

  public void setInstancesDepth(int depth) {
    _instancesDepth = depth;
    if (_instancesDepthLimited) {
      updateInstanceTree(null);
      _instances.fireMarkChanged();
    }
  }

  public void setInstancesDepth(boolean limited) {
    _instancesDepthLimited = limited;
    if (_instancesDepth < _instancesMaxDepth) {
      updateInstanceTree(null);
      _instances.fireMarkChanged();
    }
  }

  public void setApplyImmediately(boolean applyImmediately) {
    _applyImmediately = applyImmediately;
  }

  /**
   * Sets current selection, informs listeners.
   *
   * @param selection  Collection of instances to be selected.
   * @param source  The source of the selection modification request.
   */
  public void setSelection(final Collection<HepRepInstance> selection, Object source) {
    if (selection == null) {
      _selection = null;
    } else if (selection.isEmpty()) {
      _selection = Collections.emptySet();
      _instances.getSelectionModel().clearSelection();
    } else {
      _selection = Collections.newSetFromMap(new IdentityHashMap());
      _selection.addAll(selection);
      final ArrayList<TreePath>paths = new  ArrayList(selection.size());
      HepRepTreeModel.TreeVisitor v = new HepRepTreeModel.TreeVisitor() {
        public boolean visit(TreePath path) {
          Object payload = ((WiredTreeNode) path.getLastPathComponent()).getPayload();
          if (_selection.contains(payload)) paths.add(path);
          return true;
        }
      };
      v.visitPaths(_instances, 2, -1);
      _instanceTreeSelectionListenerEnabled = false;
      _instances.getSelectionModel().setSelectionPaths(paths.toArray(new TreePath[paths.size()]));
      _instanceTreeSelectionListenerEnabled = true;
    }
    fireEvent(source==null ? this : source , Event.Type.SELECTION);
  }

  /**
   * Called by picking tools to inform this model that a new set of instances has been picked.
   * The model will apply mother-daughter selection rules and update current selection.
   *
   * @param pickedInstances  Collection of newly picked instances.
   * @param source  Tool that performed the picking.
   */
  public void setPick(Collection<HepRepInstance> pickedInstances, Object source) {
    if (pickedInstances == null || pickedInstances.isEmpty()) {
      _pick = Collections.emptySet();
    } else {
      _pick = Collections.newSetFromMap(new IdentityHashMap());
      Set<HepRepInstance> parentInstances = Collections.newSetFromMap(new IdentityHashMap());
      Set<HepRepInstance> pickParentInstances = Collections.newSetFromMap(new IdentityHashMap());
      for (HepRepInstance instance : pickedInstances) {
        while (instance.getAttValue("PickParent").getBoolean()) {
          pickParentInstances.add(instance);
          instance = instance.getSuperInstance();
        }
        parentInstances.add(instance);
      }
      _pick.addAll(parentInstances);
      if (parentInstances.size() == 1) {  // one parent, add all pickParent children
        _pick.addAll(pickParentInstances);
      }
    }
    setSelection(_pick, source);
  }

  private void syncInstanceTypeTrees(Set<? extends Object> visTypes) {
    final Set<? extends Object> visibleTypes = (visTypes == null) ? _types.getMarkedObjects() : visTypes;
    HepRepTreeModel.TreeVisitor v = new HepRepTreeModel.TreeVisitor() {
      public boolean visit(WiredTreeNode node) {
        Object payload = node.getPayload();
        if (payload instanceof HepRepInstance) {
          HepRepInstance instance = (HepRepInstance)payload;
          node.setHidden( isInvisibleInstance(instance) || ! visibleTypes.contains(instance.getType()));
        }
        return true;
      }
    };
    v.visitNodes(_instances, _instancesDepthLimited ? _instancesDepth : Integer.MAX_VALUE);
  }

  private void updateTypeTree() {
    HepRepTreeModel.TreeVisitor v = new HepRepTreeModel.TreeVisitor() {
      public boolean visit(WiredTreeNode node) {
        Object payload = node.getPayload();
        node.setHidden( ((payload instanceof HepRepType) && isInvisibleType((HepRepType)payload)) || 
                              (_typesDepthLimited && node.getLevel() > _typesDepth) );
        return true;
      }
    };
    v.visitNodes(_types);
  }

  private void updateInstanceTree(Set<? extends Object> visTypes) {
    final Set<? extends Object> visibleTypes = (visTypes == null) ? _types.getMarkedObjects() : visTypes;
    HepRepTreeModel.TreeVisitor v = new HepRepTreeModel.TreeVisitor() {
      public boolean visit(WiredTreeNode node) {
        Object payload = node.getPayload();
        if (payload instanceof HepRepInstance) {
          HepRepInstance instance = (HepRepInstance)payload;
          node.setHidden(( isInvisibleInstance(instance)) || 
                                (_instancesDepthLimited && node.getLevel() > _instancesDepth) ||
                               ! visibleTypes.contains(instance.getType()));
        } else {
          node.setHidden(_instancesDepthLimited && node.getLevel() > _instancesDepth);
        }
        return true;
      }
    };
    v.visitNodes(_instances);
  }

  private boolean isInvisibleInstance(HepRepInstance instance) {
    return false;
  }

  private boolean isInvisibleType(HepRepType type) {
    return false;
  }

  private void updatePickTree(Set<? extends Object> visTypes) {
    final Set<? extends Object> visibleTypes = (visTypes == null) ? _types.getMarkedObjects() : visTypes;
    HepRepTreeModel.TreeVisitor v = new HepRepTreeModel.TreeVisitor() {
      public boolean visit(WiredTreeNode node) {
        node.setHidden(!visibleTypes.contains(node.getPayload()));
        return true;
      }
    };
    v.visitNodes(_pickTypes);
  }

  public void setCutSet(CutSet cutSet) {
    _cutSet = cutSet;
    fireEvent(this , Event.Type.CUTS);
  }


// -- Setters : ----------------------------------------------------------------

  public void setName(String name) {
    _name = name;
  }

// -- Getters : ----------------------------------------------------------------

  public String getName() {
    return _name;
  }

  /**
   * Returns the number of times this method has been called on this HepRepPanelModel.
   * On the first invocation, returns 1.
   * Used by WiredPlugin to assign names to newly created panels associated with this model.
   */
  public int getNextPanelNumber() {
    return _nPanel++;
  }

  public HepRepTreeModel getTypesModel() {return _types;}

  public HepRepTreeModel getInstancesModel() {return _instances;}

  public int getTypesDepth() {return _typesDepth;}

  public int getTypesMaxDepth() {return _typesMaxDepth;}

  public int getInstancesDepth() {return _instancesDepth;}

  public int getInstancesMaxDepth() {return _instancesMaxDepth;}

  public boolean isTypesDepthLimited() {return _typesDepthLimited;}

  public boolean isInstancesDepthLimited() {return _instancesDepthLimited;}

  public boolean isAppliedImmediately() { return _applyImmediately;}

  public Set<HepRepInstance> getSelection() {
    if (_selection == null) {
      TreeSelectionModel sm = _instances.getSelectionModel();
      if (sm.isSelectionEmpty()) {
        _selection = Collections.emptySet();
        return _selection;
      } else {
        _selection = Collections.newSetFromMap(new IdentityHashMap());
        for (TreePath path : sm.getSelectionPaths()) {
          Object payload = ((WiredTreeNode) path.getLastPathComponent()).getPayload();
          try {
            _selection.add((HepRepInstance) payload);
          } catch (ClassCastException x) {}
        }
      }
    }
    return Collections.unmodifiableSet(_selection);
  }

  public Set<HepRepInstance> getPick() {
    return Collections.unmodifiableSet(_pick);
  }

  public List<Flag> getPickableLayers() {return _pickLayers;}

  public List<String> getPickableLayerNames() {
    List out = new ArrayList(_pickLayers.size());
    for (Flag layer : _pickLayers) {
      if (layer.isSet()) out.add(layer.getName());
    }
    return out;
  }

  public HepRepTreeModel getPickableTypesModel() {return _pickTypes;}

  public Set<HepRepType> getPickableTypes() {
    return _pickTypes.getMarkedObjects();
  }

  public List<Flag> getAttCategories() {
    return _attCategories;
  }

  public CutSet getCutSet() {
    return _cutSet;
  }


// -- Handling change listeners : ----------------------------------------------

  public static class Event extends ChangeEvent {
    public enum Type {SELECTION, VISIBILITY, PICK_OPTIONS, CUTS}
    private EnumSet<Type> _types;
    private Event(Object source) {
      super(source);
      _types = null;
    }
    private Event(Object source, Type... types) {
      super(source);
      _types = EnumSet.copyOf(Arrays.asList(types));
    }
    public EnumSet<Type> getTypes() {return _types == null ? EnumSet.allOf(Type.class) : _types;}
    public boolean contains(Type type) {return _types == null ? true : _types.contains(type);}
  }

  public void addListener(ChangeListener listener) {
    _listeners.put(listener, null);
  }

  public void addListener(ChangeListener listener, Event.Type... types) {
    _listeners.put(listener, EnumSet.copyOf(Arrays.asList(types)));
  }

  public void removeListener(ChangeListener listener) {
    _listeners.remove(listener);
  }

  public void fireEvent() {
    fireEvent(this);
  }

  public void fireEvent(Object source) {
    ArrayList<ChangeListener> listeners = new ArrayList(_listeners.keySet());
    Event event = new Event(source);
    for (ChangeListener listener : listeners) listener.stateChanged(event);
  }

  public void fireEvent(Event.Type... types) {
    fireEvent(this, types);
  }

  public void fireEvent(Event.Type type) {
    fireEvent(this, type);
  }
  
  public void fireEvent(Object source, Event.Type... types) {
    ArrayList<Map.Entry<ChangeListener,EnumSet<Event.Type>>> listeners = new ArrayList(_listeners.entrySet());
    Event event = new Event(source, types);
    for (Map.Entry<ChangeListener,EnumSet<Event.Type>> entry : listeners) {
      EnumSet<Event.Type> listenerTypes = entry.getValue();
      if (listenerTypes == null) {
        entry.getKey().stateChanged(event);
      } else {
        for (Event.Type t : types) {
          if (listenerTypes.contains(t)) {
            entry.getKey().stateChanged(event);
            break;
          }
        }
      }
    }
  }

// -- Setting the model to a different HepRep object : -------------------------

  public void setRecord(HepRep heprep) {
//    System.out.println("HepRepPanelModel.setRecord, name = "+ _name +" heprep = "+ heprep);

    if (heprep == _heprep) {
      return;
    } else {
      _heprep = heprep;
    }

    if (_types != null) {
      _types.removeMarkListener(this);
      _instances.removeMarkListener(this);
      _instances.getSelectionModel().removeTreeSelectionListener(_instanceTreeSelectionListener);
      _types.saveTreeState(_keyTypeTree +"."+ _name);
      saveFlags(_keyPickTree +"."+ _name, _pickLayers);
      _pickTypes.saveTreeState(_keyPickTree +"."+ _name);
      saveFlags(_keyShowAtt +"."+ _name, _attCategories);
    }
    if (_typesMaxDepth == _typesDepth) _typesDepth = -1;
    if (_instancesMaxDepth == _instancesDepth) _instancesDepth = -1;

    if (heprep == null) {
      setDefaults();
    } else {

      _types = new HepRepTreeModel(heprep, true);
      _types.restoreTreeState(_keyTypeTree +"."+ _name);
      _types.addMarkListener(this);
      _instances = new HepRepTreeModel(heprep, false);
      _instances.addMarkListener(this);
      _instances.getSelectionModel().addTreeSelectionListener(_instanceTreeSelectionListener);

      _typesMaxDepth = _types.getMaximumDepth();
      _instancesMaxDepth = _instances.getMaximumDepth();
      if (_typesDepth == -1) _typesDepth = _typesMaxDepth;
      if (_instancesDepth == -1) _instancesDepth = _instancesMaxDepth;

      _pickLayers = restoreFlags(_keyPickTree +"."+ _name, heprep.getLayerOrder());
      _pickTypes = new HepRepTreeModel(heprep, true);
      _pickTypes.restoreTreeState(_keyPickTree +"."+ _name);
      _pickTypes.addMarkListener(new WiredTreeModel.MarkListener() {
        public void markChanged(WiredTreeModel.MarkEvent e) {
          fireEvent(Event.Type.PICK_OPTIONS);
        }
      });
      _attCategories = restoreFlags(_keyShowAtt +"."+ _name, WiredHepRepUtil.getCategories(heprep));

      setSelection(null, this);
    }
  }

// -- Local methods : ----------------------------------------------------------

  private void saveFlags(String key, Collection<Flag> flags) {
    List<String> unsetFlags = new ArrayList();
    for (Flag f : flags) {
      if (!f.isSet()) unsetFlags.add(f.getName());
    }
    Properties user = Application.getApplication().getUserProperties();
    PropertyUtilities.setStringCollection(user, key, unsetFlags);
  }

  private ArrayList<Flag> restoreFlags(String key, Collection<String> flagNames) {
    Properties user = Application.getApplication().getUserProperties();
    HashSet<String> unsetFlags = new HashSet(PropertyUtilities.getStringCollection(user, key, Collections.EMPTY_SET));
    ArrayList<Flag> out = new ArrayList(flagNames.size());
    for (String name : flagNames) {
      out.add(new Flag(name, !unsetFlags.contains(name)));
    }
    return out;
  }

// -----------------------------------------------------------------------------
}
