// Copyright 2004-2011, FreeHEP.
package hep.wired.heprep.interaction;

import hep.graphics.heprep.HepRepInstance;
import hep.wired.heprep.graphicspanel.HepRepPanel;
import hep.wired.heprep.util.InsideRectanglePathConstructor;
import hep.wired.heprep.util.PathGraphics2D;
import hep.wired.image.WiredBaseImage;
import hep.wired.interaction.DragRectangleToScale;
import hep.wired.services.RecordPlot;
import hep.wired.util.XYZindices;

import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;
import java.util.List;
import java.util.Set;

import javax.swing.Icon;

import org.freehep.application.Application;
import org.freehep.graphicsio.PathConstructor;

/**
 * InteractionHandler that allows user to pick objects by drawing a rectangle on the plot.
 * 
 * @author Mark Donszelmann
 * @version $Id: DragRectangleToPick.java 13838 2011-05-17 22:57:22Z onoprien $
 */
public class DragRectangleToPick extends PickInteractionHandler implements PickHandler, XYZindices {

// -- Private parts : ----------------------------------------------------------

  private static final int cursorSize = 32;
  private double x0;
  private double y0;
  private double x;
  private double y;
  Rectangle2D rectangle;

// -- Construction : -----------------------------------------------------------

  public DragRectangleToPick() {
    super("Drag Rectangle to Pick", false);
  }

// -- Implementing PickHandler : -----------------------------------------------

  public void update(RecordPlot plot) {
    if (rectangle != null) {
      selectInstances(plot);
      plot.drawShape(rectangle);
    }
  }

  public boolean isRegionZoomable() {
    return rectangle != null;
  }

  public void zoomIntoRegion(RecordPlot plot) {
    if (!isRegionZoomable()) return;
    DragRectangleToScale.translateAndScale(plot, rectangle, true);
//    clearShape(plot);
  }

// -- Implementing InteractionHandler : ----------------------------------------

  public void setSelected(RecordPlot plot, boolean selected) {
    super.setSelected(plot, selected);
    if (selected && rectangle != null) plot.drawShape(rectangle);
  }

  public void setSize(RecordPlot plot, int width, int height) {
    if ((getWidth() > 0) && (getHeight() > 0)) {
      double sx = (double) width / getWidth();
      double sy = (double) height / getHeight();
      x0 = (int) (x0 * sx);
      y0 = (int) (y0 * sy);
      defineShape(plot, (int) (x * sx), (int) (y * sy));
//      selectInstances(plot);
    }
    super.setSize(plot, width, height);
  }

  public Icon getIcon(int size) {
    return WiredBaseImage.getIcon("Rectangle%w", size);
  }

  public String getDescription() {
    return "Click and drag a rectangle to pick objects.";
  }

  public void changeCursor(RecordPlot plot, InputEvent event) {
    plot.setCursor(WiredBaseImage.getBestCursor("RectangleCursor%w", cursorSize, cursorSize));
  }

  public void mouseButton1Clicked(RecordPlot plot, MouseEvent event) {
    startShape(0.,0.);
    plot.drawShape(null);
    setSelection(plot, null);
  }

  public void mouseButton1DragStarted(RecordPlot plot, MouseEvent event) {
    startShape(event.getX(), event.getY());
    defineShape(plot, event.getX(), event.getY());
    selectInstances(plot);
  }

  public void mouseButton1Dragged(RecordPlot plot, MouseEvent event) {
    defineShape(plot, event.getX(), event.getY());
    if (pickWhileDragging) {
      selectInstances(plot);
    }
  }

  public void mouseButton1DragEnded(RecordPlot plot, MouseEvent event) {
    defineShape(plot, event.getX(), event.getY());
    selectInstances(plot);
//    clearShape(plot);
  }

  public void mouseEntered(RecordPlot plot, MouseEvent event) {
//        info().setInteractionHandler(this, plot);
//        info().setSelection(true);
    plot.requestFocusInWindow();
    changeCursor(plot, event);
  }

//  public void mouseWheelMoved(RecordPlot plot, MouseWheelEvent event) {
//    super.mouseWheelMoved(plot, event);
//
//    // update selection rectangle
//    double s = getLastScale();
//    double w2 = plot.getWidth() / 2;
//    double h2 = plot.getHeight() / 2;
//    startShape((x0 - w2) * s + w2, (y0 - h2) * s + h2);
//    defineShape(plot, (x - w2) * s + w2, (y - h2) * s + h2);
//
//    selectInstances(plot);
//  }

// -- Local methods : ----------------------------------------------------------

  protected void clearShape(RecordPlot plot) {
    startShape(0, 0);
    x = 0;
    y = 0;
    plot.drawShape(null);
  }

  private void startShape(double x0, double y0) {
    this.x0 = x0;
    this.y0 = y0;
    rectangle = null;
  }

  private void defineShape(RecordPlot plot, double x, double y) {
    this.x = x;
    this.y = y;

    double w = Math.abs(x - x0);
    double h = Math.abs(y - y0);
    if ((w != 0) && (h != 0)) {
      rectangle = new Rectangle2D.Double(Math.min(x, x0), Math.min(y, y0), w, h);
    } else {
      rectangle = null;
    }

    plot.drawShape(rectangle);
  }

  private void selectInstances(RecordPlot plot) {
    pickedInstances.clear();
    selectedInstances.clear();
    if (rectangle != null) {
      InsideRectanglePathConstructor pathConstructor = new InsideRectanglePathConstructor(rectangle) {
        protected void lineFound(Line2D line) {
          if (!selectedInstances.contains(currentInstance)) {
            pickedInstances.add(currentInstance);
            selectedInstances.add(currentInstance);
          }
        }
      };
      draw(plot, pathConstructor);
    }
    setSelection(plot, pickedInstances);
  }

  private void draw(RecordPlot plot, PathConstructor pathConstructor) {

    HepRepPanel panel = (HepRepPanel) plot.getGraphicsPanel();
    panel.setHepRepInstanceListener(new HepRepPanel.InstanceListener() {
      public void currentHepRepInstance(HepRepInstance instance) {
        currentInstance = instance;
      }
    });

    PathGraphics2D pathGraphics = new PathGraphics2D(pathConstructor);
    List layers = panel.getModel().getPickableLayerNames();
    Set types = panel.getModel().getPickableTypes();
    panel.draw(pathGraphics, true, layers.isEmpty() ? null : layers, types.isEmpty() ? null : types, true);

    panel.setHepRepInstanceListener(null);
  }


// -- Overriding Object : ------------------------------------------------------

  public String toString() {
    return "Drag Rectangle to Pick";
  }
}
