// Copyright 2004-2011, FreeHEP.
package hep.wired.heprep.interaction;

import hep.graphics.heprep.HepRepInstance;
import hep.wired.heprep.graphicspanel.HepRepPanel;
import hep.wired.heprep.util.Arrow;
import hep.wired.heprep.util.NearestPointPathConstructor;
import hep.wired.heprep.util.PathGraphics2D;
import hep.wired.image.WiredBaseImage;
import hep.wired.services.GraphicsPanel;
import hep.wired.services.RecordPlot;

import java.awt.Cursor;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Point2D;
import java.util.List;
import java.util.Set;

import javax.swing.Icon;

import org.freehep.application.Application;
import org.freehep.graphicsio.PathConstructor;

/**
 * Hover mouse near to object to show info.
 * 
 * @author Mark Donszelmann
 * @version $Id: HoverToPick.java 14093 2013-01-09 21:57:40Z onoprien $
 */
public class HoverToPick extends PickInteractionHandler {
  
// -- Private parts : ----------------------------------------------------------

  private int x;
  private int y;
  private boolean fixed = false;
  private Point2D point;
  private double minimumDistanceSq;
  
// -- Construction : -----------------------------------------------------------

  public HoverToPick() {
    super("Hover near Object to Pick", true);
  }

// -- Getters : ----------------------------------------------------------------

  public Icon getIcon(int size) {
    return WiredBaseImage.getIcon("PickInfo%w", size);
  }

  public boolean isRegionZoomable() {
    return false;
  }

  public String getDescription() {
    return "Hover over or near an object to pick.";
  }

// -- Implementing PickHandler : -----------------------------------------------

  /**
   * Performs picking action based on current cursor location if possible,
   * decorates the plot and informs it of newly picked objects.
   */
  public void update(RecordPlot plot) {
    if ((plot.getWidth() != 0) && (plot.getHeight() != 0)) {
      defineNearestPoint(plot, x, y);
      defineShape(plot);
    }
  }

  public void zoomIntoRegion(RecordPlot plot) {
  }

  public void translateToPicked(RecordPlot plot) {
    super.translateToPicked(plot);
    fixed = false;
  }

// -- Local methods that support picking functionality : -----------------------

  protected void clearShape(RecordPlot plot) {
    x = 0;
    y = 0;
//    try {
//      ((HepRepPanel) plot.getGraphicsPanel()).getModel().setPick(null, this);
//    } catch (ClassCastException xx) {
//    }
    plot.drawShape(null);
  }

  /** Memorizes the point and clears internal lists of selected and picked instances. */
  private void definePoint(RecordPlot plot, int x, int y) {
    this.x = x;
    this.y = y;
    point = new Point2D.Double(x, y);
    pickedInstances.clear();
    selectedInstances.clear();
  }

  private void defineNearestPoint(RecordPlot plot, int x, int y) {
    definePoint(plot, x, y);

    minimumDistanceSq = 0;
    NearestPointPathConstructor pathConstructor = new NearestPointPathConstructor(point) {
      protected void pointFound(Point2D p, double distanceSq, boolean closer) {
        if (closer) {
          minimumDistanceSq = distanceSq;
          pickedInstances.clear();
        }
        pickedInstances.add(currentInstance);
      }
    };
    draw(plot, pathConstructor);

    Point2D nearestPoint = pathConstructor.getNearestPoint();
    if (nearestPoint != null) {
      point = nearestPoint;
    }
  }

  /** Updates selection in the plot model and draws a shape on the plot. */
  private void defineShape(RecordPlot plot) {
    try {
      ((HepRepPanel) plot.getGraphicsPanel()).getModel().setPick(pickedInstances, this);
    } catch (ClassCastException xx) {}

    // Farthest corner
    Point2D corner = new Point2D.Double((point.getX() > plot.getX()
            + plot.getWidth() / 2) ? plot.getX() : plot.getX()
            + plot.getWidth(), (point.getY() > plot.getY()
            + plot.getHeight() / 2) ? plot.getY() : plot.getY()
            + plot.getHeight());

    GeneralPath path = new GeneralPath(new Arrow(corner, point, 20, 16, 70, false));
    if (!fixed && pickWhileDragging) {
      double distance = Math.max(Math.sqrt(minimumDistanceSq), 40);
      path.append(new Ellipse2D.Double(point.getX() - distance, point.getY() - distance, distance * 2, distance * 2), false);
    }
    plot.drawShape(path);
  }

  private void draw(RecordPlot plot, PathConstructor pathConstructor) {

    HepRepPanel panel = (HepRepPanel) plot.getGraphicsPanel();
    panel.setHepRepInstanceListener(new HepRepPanel.InstanceListener() {
      public void currentHepRepInstance(HepRepInstance instance) {
        currentInstance = instance;
      }
    });

    PathGraphics2D pathGraphics = new PathGraphics2D(pathConstructor);
    List layers = panel.getModel().getPickableLayerNames();
    Set types = panel.getModel().getPickableTypes();
    panel.draw(pathGraphics, true, layers.isEmpty() ? null : layers, types.isEmpty() ? null : types, true);

    panel.setHepRepInstanceListener(null);
  }

// -- InteractionHandler hooks : -----------------------------------------------

  public void setSelected(RecordPlot plot, boolean selected) {
    super.setSelected(plot, selected);
    if (selected) {
      try {
       fixed = !((HepRepPanel) plot.getGraphicsPanel()).getModel().getSelection().isEmpty();
      } catch (ClassCastException x) {}
    }
  }

  public void setRecord(RecordPlot plot, Object record) {
    fixed = false;
  }

//  public void reset(RecordPlot plot, InputEvent event) {
//    Application.getApplication().setStatusMessage(getDescription());
//    fixed = false;
//    clearShape(plot);
//  }

  public void changeCursor(RecordPlot plot, InputEvent event) {
    plot.setCursor(Cursor.getDefaultCursor());
  }

// -- Responding to mouse events : ---------------------------------------------

  public void mouseMoved(RecordPlot plot, MouseEvent event) {
    super.mouseMoved(plot, event);
    changeCursor(plot, event);
    if (!fixed) {
      if (pickWhileDragging) {
        defineNearestPoint(plot, event.getX(), event.getY());
      } else {
        definePoint(plot, event.getX(), event.getY());
      }
      defineShape(plot);
    }
  }

  public void mouseButton1Clicked(RecordPlot plot, MouseEvent event) {
    fixed = pickWhileDragging ? !fixed : true;
    defineNearestPoint(plot, event.getX(), event.getY());
    defineShape(plot);
    clearShape(plot);
  }

  public void mouseEntered(RecordPlot plot, MouseEvent event) {
    if (!fixed) {
//            info().setInteractionHandler(this, plot);
      plot.requestFocusInWindow();
      if (pickWhileDragging) {
        defineNearestPoint(plot, event.getX(), event.getY());
      } else {
        definePoint(plot, event.getX(), event.getY());
      }
      defineShape(plot);
    }
  }

  public void mouseExited(RecordPlot plot, MouseEvent event) {
    if (!fixed) {
      clearShape(plot);
      setSelection(plot, null);
    }
  }

  public void mouseWheelMoved(RecordPlot plot, MouseWheelEvent event) {
    super.mouseWheelMoved(plot, event);
    // FIXME WIRED-382, we may want to stick to the same object...
    fixed = false;

    mouseMoved(plot, event);
  }

// -- Overriding Object : ------------------------------------------------------

  public String toString() {
    return "Hover to Pick";
  }
}
