// Copyright 2004-2006, FreeHEP.
package hep.wired.heprep.projection;

import hep.wired.heprep.services.Projection;
import hep.wired.services.Feature;
import hep.wired.services.ViewPort;

import org.freehep.xml.io.XMLIOManager;

/**
 * Base class for projections.
 * 
 * @author Mark Donszelmann
 * @version $Id: AbstractProjection.java 8598 2006-08-14 20:38:02Z duns $
 */

public abstract class AbstractProjection implements Projection {

    // FIXME: pickup from defaults?
    public static final String defaultProjections = 
        "ParallelProjection.XY:ParallelProjection.ZX:ParallelProjection.ZY:ParallelProjection.XZ:ParallelProjection.YZ:"+
        "FlatProjection:Rho-Z:YX-FishEye:Rho-Z-FishEye"; //":YX-Clock";

    private String name;

    private String id;

    private transient double[] uvw = new double[3];

    private transient ViewPort viewPort;

    /**
     * Creates a projection with the classname as name.
     */
    public AbstractProjection() {
        this(null);
    }

    /**
     * Creates a named projection.
     */
    public AbstractProjection(String name) {
        if (name == null) {
            String className = getClass().getName();
            int dot = className.lastIndexOf(".");
            className = dot < 0 ? className : className.substring(dot + 1);
            this.id = className;
            this.name = className.replace('$', '-');
        } else {
            this.name = name;
            this.id = name;
        }
    }

    public String getID() {
        return id;
    }

    public String getName() {
        return name;
    }

    public String getFormula() {
        return "(u,v,w) = " + getName() + "(x,y,z)";
    }

    public boolean supports(Class featureClass) {
        return getFeature(featureClass) != null;
    }

    /**
     * Returns feature if supported by this projection.
     */
    public Feature getFeature(Class featureClass) {
        return (featureClass.isAssignableFrom(this.getClass())) ? (Feature) this
                : null;
    }

    public void setViewPort(ViewPort viewPort) {
        this.viewPort = viewPort;
    }

    public ViewPort getViewPort() {
        return viewPort;
    }

    /**
     * Implements this method by a call to transform([]), followed by the
     * viewPort.
     */
    public double[] transform(ViewPort viewPort, double[] xyz) {
        xyz = transform(xyz);
        xyz = viewPort.transform(xyz);
        return xyz;
    }

    public abstract double[] transform(double[] xyz);

    /**
     * Implements this method by a call to transform([][], n), followed by
     * running all points through the viewPort. Concrete projections should
     * probably override this method with something more efficient.
     */
    public double[][] transform(ViewPort viewPort, double[][] xyz, int n) {
        xyz = transform(xyz, n);
        xyz = viewPort.transform(xyz, n);
        return xyz;
    }

    /**
     * Implements this method in terms of double[] transform(double[] xyz). Not
     * a very efficient implementation, so concrete projections should probably
     * override this method with something better.
     */
    public double[][] transform(double[][] xyz, int n) {
        for (int i = 0; i < n; i++) {
            uvw[U] = xyz[X][i];
            uvw[V] = xyz[Y][i];
            uvw[W] = xyz[Z][i];
            uvw = transform(uvw);
            xyz[U][i] = uvw[U];
            xyz[V][i] = uvw[V];
            xyz[W][i] = uvw[W];
        }
        return xyz;
    }

    public abstract double[] deltaTransform(double[] xyz);

//    public abstract double[] inverseTransform(double[] uvw)
//            throws UnsupportedOperationException;

//    public abstract double[] inverseDeltaTransform(double[] uvw)
//            throws UnsupportedOperationException;

    public abstract Projection copy();

    /**
     * Returns name of projection
     */
    public String toString() {
        return getName();
    }

    //
    // XMLIO
    //
    public void save(XMLIOManager xmlioManager, org.jdom.Element nodeEl) {
        nodeEl.setAttribute("name", getName());
    }

    public void restore(XMLIOManager xmlioManager, org.jdom.Element nodeEl) {
        name = nodeEl.getAttributeValue("name");
    }
}
