// Copyright 2004-2006, FreeHEP.
package hep.wired.heprep.projection;

import hep.wired.feature.Resetable;
import hep.wired.feature.Rotateable2D;
import hep.wired.feature.Rotateable3D;
import hep.wired.feature.Transformable2D;
import hep.wired.feature.Transformable3D;
import hep.wired.heprep.services.InverseProjection;
import hep.wired.heprep.services.Projection;
import hep.wired.services.ViewPort;
import hep.wired.util.Matrix3D;

import java.awt.geom.NoninvertibleTransformException;
import java.util.Iterator;

import org.freehep.xml.io.XMLIOManager;
import org.jdom.Element;

/**
 * 3D parallel projection of (X,Y,Z) into (U,V,W).
 * 
 * @author Mark Donszelmann
 * @version $Id: ParallelProjection.java 8598 2006-08-14 20:38:02Z duns $
 */

public abstract class ParallelProjection extends AbstractProjection implements InverseProjection,
        Rotateable2D, Transformable2D, Rotateable3D, Transformable3D, Resetable {

    // FIXME WIRED-242, take from preferences
    private final static double defaultScaleFactor = 0.001;

    protected Matrix3D matrix;

    protected Matrix3D resetMatrix;

    public static class XY extends ParallelProjection {
        public XY() {
            super();
        }

        protected XY(Matrix3D matrix, Matrix3D resetMatrix) {
            super(matrix, resetMatrix);
        }

        public Projection copy() {
            return new XY(matrix, resetMatrix);
        }
    }

    public static class ZX extends ParallelProjection {
        public ZX() {
            super(new Matrix3D(0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0));
        }

        protected ZX(Matrix3D matrix, Matrix3D resetMatrix) {
            super(matrix, resetMatrix);
        }

        public Projection copy() {
            return new ZX(matrix, resetMatrix);
        }
    }

    public static class ZY extends ParallelProjection {
        public ZY() {
            super(new Matrix3D(0, 0, -1, 0, 1, 0, 1, 0, 0, 0, 0, 0));
        }

        protected ZY(Matrix3D matrix, Matrix3D resetMatrix) {
            super(matrix, resetMatrix);
        }

        public Projection copy() {
            return new ZY(matrix, resetMatrix);
        }
    }

    public static class XZ extends ParallelProjection {
        public XZ() {
            super(new Matrix3D(1, 0, 0, 0, 0, -1, 0, 1, 0, 0, 0, 0));
        }

        protected XZ(Matrix3D matrix, Matrix3D resetMatrix) {
            super(matrix, resetMatrix);
        }

        public Projection copy() {
            return new XZ(matrix, resetMatrix);
        }
    }

    public static class YZ extends ParallelProjection {
        public YZ() {
            super(new Matrix3D(0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0));
        }

        protected YZ(Matrix3D matrix, Matrix3D resetMatrix) {
            super(matrix, resetMatrix);
        }

        public Projection copy() {
            return new YZ(matrix, resetMatrix);
        }
    }

    /**
     * Creates a 3D parallel projection.
     */
    protected ParallelProjection() {
        this(new Matrix3D());
    }

    protected ParallelProjection(Matrix3D matrix) {
        this(matrix, matrix);
        this.matrix.scale(defaultScaleFactor, defaultScaleFactor,
                defaultScaleFactor);
        this.resetMatrix.scale(defaultScaleFactor, defaultScaleFactor,
                defaultScaleFactor);
    }

    /**
     * Used for cloning
     */
    protected ParallelProjection(Matrix3D matrix, Matrix3D resetMatrix) {
        super();
        this.matrix = (Matrix3D) matrix.clone();
        this.resetMatrix = (Matrix3D) resetMatrix.clone();
    }

    public String getFormula() {
        return "(u,v,w) = " + "m00*x + m01*y + m02*z + m03, "
                + "m10*x + m11*y + m12*z + m13, "
                + "m20*x + m21*y + m22*z + m23";
    }

    public double[] transform(double[] xyz) {
        return matrix.transform(xyz, false);
    }

    public double[][] transform(double[][] xyz, int n) {
        return matrix.transform(xyz, n, false);
    }

    public double[] deltaTransform(double[] xyz) {
        return matrix.transform(xyz, true);
    }

    public double[] inverseTransform(double[] uvw)
            throws UnsupportedOperationException {
        try {
            Matrix3D inverse = matrix.createInverse();
            return inverse.transform(uvw, false);
        } catch (NoninvertibleTransformException e) {
            throw new UnsupportedOperationException(e.getMessage());
        }
    }

    public double[] inverseDeltaTransform(double[] uvw)
            throws UnsupportedOperationException {
        try {
            Matrix3D inverse = matrix.createInverse();
            return inverse.transform(uvw, true);
        } catch (NoninvertibleTransformException e) {
            throw new UnsupportedOperationException(e.getMessage());
        }
    }

    public String toString() {
        return super.toString() + " : \n" + matrix.toString();
    }

    // Rotateable3D
    public void rotate(double theta, double nx, double ny, double nz) {
        matrix.rotate(theta, nx, ny, nz);
    }

    // Rotateable2D
    public void rotate(double theta) {
        rotate(theta, 0, 0, 1);
    }

    // Translateable
    public double[] getModelTranslation(double[] uvw, ViewPort viewPort) {
        uvw = viewPort.inverseDeltaTransform(uvw);
        return inverseDeltaTransform(uvw);
    }

    public double[] getScreenTranslation(double[] xyz, ViewPort viewPort) {
        xyz = deltaTransform(xyz);
        return viewPort.deltaTransform(xyz);
    }

    // Transformable3D
    public void scale(double s) {
        scale(s, s, s);
    }

    public void setScale(double s) {
        setScale(s, s, s);
    }

    public void scale(double sx, double sy, double sz) {
        matrix.scale(sx, sy, sz);
    }

    public void setScale(double sx, double sy, double sz) {
        double[] s = getScale();
        matrix.scale(sx / s[X], sy / s[Y], sz / s[Z]);
    }

    public double[] getScale() {
        return new double[] { matrix.getScaleX(), matrix.getScaleY(),
                matrix.getScaleZ() };
    }

    public void translate(double tx, double ty, double tz) {
        matrix.modelTranslate(tx, ty, tz);
    }

    public void setTranslate(double tx, double ty, double tz) {
        double[] t = getTranslate();
        matrix.modelTranslate(tx - t[X], ty - t[Y], tz - t[Z]);
    }

    public double[] getTranslate() {
        return inverseDeltaTransform(new double[] { matrix.getTranslateX(),
                matrix.getTranslateY(), matrix.getTranslateZ() });
    }

    // Transformable2D
    public void scale(double sx, double sy) {
        scale(sx, sy, 1);
    }

    public void setScale(double sx, double sy) {
        setScale(sx, sy, matrix.getScaleZ());
    }

    /**
     * Shears the viewpoint. Preconcatenates with the matrix:
     * 
     * <pre>
     *    [   1               shx             0               0       ]
     *    [   shy             1               0               0       ]
     *    [   0               0               1               0       ]
     *    [   0               0               0               1       ]
     * </pre>
     */
    public void shear(double shx, double shy) {
        matrix.shear(shx, shy);
    }

    public void translate(double tx, double ty) {
        translate(tx, ty, 0);
    }

    public void setTranslate(double tx, double ty) {
        setTranslate(tx, ty, matrix.getTranslateZ());
    }

    // Resetable
    public Object reset(Object newState) {
        Object oldState = matrix.clone();
        matrix = (newState == null) ? (Matrix3D) resetMatrix.clone()
                : (Matrix3D) newState;
        return oldState;
    }

    //
    // XMLIO
    //
    public void save(XMLIOManager xmlioManager, org.jdom.Element nodeEl) {
        super.save(xmlioManager, nodeEl);
        nodeEl.addContent(xmlioManager.save(matrix));
        nodeEl.addContent(xmlioManager.save(resetMatrix));
    }

    public void restore(XMLIOManager xmlioManager, org.jdom.Element nodeEl) {
        super.restore(xmlioManager, nodeEl);
        Iterator i = nodeEl.getChildren().iterator();
        matrix = (Matrix3D) xmlioManager.restore((Element) i.next());
        resetMatrix = (Matrix3D) xmlioManager.restore((Element) i.next());
    }
}
