// Copyright 2004-2006, FreeHEP.
package hep.wired.heprep.projection;

import hep.wired.feature.Resetable;
import hep.wired.heprep.services.InverseProjection;
import hep.wired.heprep.services.Projection;
import hep.wired.variable.BooleanVariable;
import hep.wired.variable.DoubleVariable;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * Projects R : sqrt(x^2 + y^2) and Z.
 * 
 * @author Mark Donszelmann
 * @version $Id: RhoZProjection.java 8598 2006-08-14 20:38:02Z duns $
 */

public class RhoZProjection extends VariableProjection implements InverseProjection, Resetable {

    private static final double PI2 = Math.PI * 2;

    private double[] uvw = new double[3];

    private DoubleVariable varPhi0;

    private double phi0, resetPhi0;

    private BooleanVariable varSplitPhi;

    private boolean splitPhi, resetSplitPhi;

    // FIXME WIRED-314, we need phi0 parameter here
    public RhoZProjection() {
        this(0, 0, true, true);
    }

    private RhoZProjection(double phi0, double resetPhi0, boolean splitPhi,
            boolean resetSplitPhi) {
        super("RhoZProjection");
        varPhi0 = addVariable("phi0", 0.0, 180, phi0, "degrees",
                "Angle at which to slice detector into two halves.");
        varSplitPhi = addVariable("splitPhi", splitPhi,
                "Slice detector into two halves.");

        varSplitPhi.addChangeListener(new ChangeListener() {
            public void stateChanged(ChangeEvent e) {
                varPhi0.setEnabled(varSplitPhi.getBooleanVariable());
            }
        });
        varPhi0.setEnabled(splitPhi);

        this.phi0 = phi0;
        this.resetPhi0 = resetPhi0;
        this.splitPhi = splitPhi;
        this.resetSplitPhi = resetSplitPhi;
    }

    public String getFormula() {
        return "(u,v,w) = " + "z, " + "sqrt(x^2 + y^2), " + "0";
    }

    public double[] transform(double[] xyz) {
        double phi = Math.atan2(xyz[Y], xyz[X]);
        int sign = getSign(phi);

        xyz = transform(xyz, sign);
        return xyz;
    }

    public double[][] transform(double[][] xyz, int n) {
        if (n == 0)
            return xyz;

        int sign = 0;
        for (int i = 0; i < n; i++) {
            if (xyz[X][i] != 0) {
                double phi = Math.atan2(xyz[Y][i], xyz[X][i]);
                sign += getSign(phi);
            }
        }
        sign = (sign >= 0) ? +1 : -1;

        for (int i = 0; i < n; i++) {
            uvw[U] = xyz[X][i];
            uvw[V] = xyz[Y][i];
            uvw[W] = xyz[Z][i];
            uvw = transform(uvw, sign);
            xyz[U][i] = uvw[U];
            xyz[V][i] = uvw[V];
            xyz[W][i] = uvw[W];
        }
        return xyz;
    }

    private int getSign(double phi) {
        phi = (phi + PI2) % PI2; // make phi in [0, 2PI]
        phi = phi - Math.toRadians(phi0); // subtract phi0
        phi = (phi + PI2) % PI2; // make phi in [0, 2PI]
        return phi <= Math.PI ? +1 : -1;
    }

    private double[] transform(double[] xyz, int sign) {
        double x = xyz[X];
        double y = xyz[Y];
        double z = xyz[Z];

        double rho = Math.sqrt((x * x) + (y * y));

        xyz[U] = z;
        xyz[V] = (splitPhi) ? sign * rho : rho;
        xyz[W] = 1;

        return xyz;
    }

    public double[] deltaTransform(double[] xyz) {
        return transform(xyz);
    }

    public double[] inverseTransform(double[] uvw) throws UnsupportedOperationException {
        double d = Math.sqrt((uvw[V] * uvw[V]) / 2);

        uvw[Z] = uvw[U];
        uvw[X] = d;
        uvw[Y] = d;
        return uvw;
    }

    public double[] inverseDeltaTransform(double[] uvw) throws UnsupportedOperationException {
        return inverseTransform(uvw);
    }
    
    public Projection copy() {
        return new RhoZProjection(phi0, resetPhi0, splitPhi, resetSplitPhi);
    }

    // Resetable
    public Object reset(Object newState) {
        Object oldState = new Object[] { new Double(phi0),
                new Boolean(splitPhi) };
        if (newState == null) {
            phi0 = resetPhi0;
            splitPhi = resetSplitPhi;
        } else {
            Object[] array = (Object[]) newState;
            phi0 = ((Double) array[0]).doubleValue();
            splitPhi = ((Boolean) array[1]).booleanValue();

        }
        varPhi0.setEnabled(splitPhi);
        return oldState;
    }
}
