// Copyright 2004-2005, FreeHEP.
package hep.wired.heprep.projection;

import hep.wired.heprep.feature.NamedVariables;
import hep.wired.variable.BooleanVariable;
import hep.wired.variable.DoubleVariable;
import hep.wired.variable.IntegerVariable;
import hep.wired.variable.NumberVariable;
import hep.wired.variable.Variable;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.freehep.xml.io.XMLIOManager;
import org.jdom.Element;

/**
 * Base class for projections based on named variables.
 * 
 * @author Mark Donszelmann
 * @version $Id: VariableProjection.java 8598 2006-08-14 20:38:02Z duns $
 */

public abstract class VariableProjection extends AbstractProjection implements
        NamedVariables {

    private String name;

    private Map/* <String, Variable> */variables = new HashMap();

    /**
     * Creates a named projection.
     */
    public VariableProjection(String name) {
        this.name = name;
    }

    public String getName() {
        return name;
    }

    /**
     * Adds a named variable with minimum, maximum and default value.
     */
    protected DoubleVariable addVariable(String name, double min, double max,
            double def, String unit, String description) {
        DoubleVariable variable = new DoubleVariable(name, this, min, max, def,
                unit, description);
        variables.put(name, variable);
        return variable;
    }

    /**
     * Adds a named variable with minimum, maximum and default value.
     */
    protected IntegerVariable addVariable(String name, int min, int max,
            int def, String unit, String description) {
        IntegerVariable variable = new IntegerVariable(name, this, min, max,
                def, unit, description);
        variables.put(name, variable);
        return variable;
    }

    /**
     * Adds a named variable with default value.
     */
    protected BooleanVariable addVariable(String name, boolean def,
            String description) {
        BooleanVariable variable = new BooleanVariable(name, this, def,
                description);
        variables.put(name, variable);
        return variable;
    }

    /**
     * Return true if variable with name exists.
     */
    public boolean exists(String name) {
        return variables.containsKey(name);
    }

    /**
     * Return set of variable names
     */
    public Set/* <String> */getVariableNames() {
        return variables.keySet();
    }

    /**
     * Returns a list of all variables
     */
    public Collection/* <Variable> */getVariables() {
        return variables.values();
    }

    /**
     * Return the unit, or null;
     */
    public String getUnit(String name) {
        Variable var = (Variable) variables.get(name);
        return (var != null) ? var.getUnit() : null;
    }

    /**
     * Return the description
     */
    public String getDescription(String name) {
        Variable var = (Variable) variables.get(name);
        return (var != null) ? var.getDescription() : null;
    }

    public double getDoubleVariable(String name) {
        NumberVariable var = (NumberVariable) variables.get(name);
        if (var == null)
            throw new IllegalArgumentException("Variable with name '" + name
                    + "' does not exist.");

        return (var instanceof DoubleVariable) ? ((DoubleVariable) var)
                .getDoubleVariable() : ((IntegerVariable) var)
                .getIntegerVariable();
    }

    public int getIntegerVariable(String name) {
        NumberVariable var = (NumberVariable) variables.get(name);
        if (var == null)
            throw new IllegalArgumentException("Variable with name '" + name
                    + "' does not exist.");

        return (var instanceof DoubleVariable) ? (int) ((DoubleVariable) var)
                .getDoubleVariable() : ((IntegerVariable) var)
                .getIntegerVariable();
    }

    public boolean getBooleanVariable(String name) {
        BooleanVariable var = (BooleanVariable) variables.get(name);
        if (var == null)
            throw new IllegalArgumentException("Variable with name '" + name
                    + "' does not exist.");

        return var.getBooleanVariable();
    }

    /**
     * Sets the value of a named variable.
     */
    public void setVariable(String name, double value) {
        DoubleVariable var = (DoubleVariable) variables.get(name);
        if (var == null)
            throw new IllegalArgumentException("Variable with name '" + name
                    + "' does not exist.");

        var.setVariable(value);
    }

    /**
     * Sets the value of a named variable.
     */
    public void setVariable(String name, int value) {
        NumberVariable var = (NumberVariable) variables.get(name);
        if (var == null)
            throw new IllegalArgumentException("Variable with name '" + name
                    + "' does not exist.");

        if (var instanceof DoubleVariable) {
            ((DoubleVariable) var).setVariable(value);
        } else {
            ((IntegerVariable) var).setVariable(value);
        }
    }

    /**
     * Sets the value of a named variable.
     */
    public void setVariable(String name, boolean value) {
        BooleanVariable var = (BooleanVariable) variables.get(name);
        if (var == null)
            throw new IllegalArgumentException("Variable with name '" + name
                    + "' does not exist.");

        var.setVariable(value);
    }

    public void multiplyVariable(String name, double factor) {
        setVariable(name, getDoubleVariable(name) * factor);
    }

    public void multiplyVariable(String name, int factor) {
        if (variables.get(name) instanceof IntegerVariable) {
            setVariable(name, getIntegerVariable(name) * factor);
        } else {
            setVariable(name, getDoubleVariable(name) * factor);
        }
    }

    public void addVariable(String name, double add) {
        setVariable(name, getDoubleVariable(name) + add);
    }

    public void addVariable(String name, int add) {
        if (variables.get(name) instanceof IntegerVariable) {
            setVariable(name, getIntegerVariable(name) + add);
        } else {
            setVariable(name, getDoubleVariable(name) + add);
        }
    }

    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append(super.toString());
        sb.append(" - [");
        for (Iterator i = variables.entrySet().iterator(); i.hasNext();) {
            Map.Entry entry = (Map.Entry) i.next();
            String name = (String) entry.getKey();
            Variable variable = (Variable) entry.getValue();

            sb.append(name);
            sb.append("=");
            sb.append(variable.toString());
            sb.append(variable.getUnit() != null ? variable.getUnit() : "");
            if (i.hasNext())
                sb.append(", ");
        }
        sb.append("]");
        return sb.toString();
    }

    //
    // XMLIO
    //
    public void save(XMLIOManager xmlioManager, org.jdom.Element nodeEl) {
        super.save(xmlioManager, nodeEl);
        for (Iterator i = variables.entrySet().iterator(); i.hasNext();) {
            Map.Entry entry = (Map.Entry) i.next();
            String name = (String) entry.getKey();
            Variable variable = (Variable) entry.getValue();

            // save definition
            Element element = xmlioManager.save(variable);
            nodeEl.addContent(element);

            // save value
            Element value = new Element(name);
            variable.saveValue(value);
            element.addContent(value);
        }
    }

    public void restore(XMLIOManager xmlioManager, org.jdom.Element nodeEl) {
        super.restore(xmlioManager, nodeEl);
        for (Iterator i = nodeEl.getChildren().iterator(); i.hasNext();) {
            Element element = (Element) i.next();
            Variable variable = (Variable) xmlioManager.restore(element);
            variables.put(name, variable);

            Element value = (Element) element.getChildren().iterator().next();
            name = value.getAttributeValue("name");
            variable.restoreValue(value);
        }
    }
}
