// Copyright 2004-2005, FreeHEP.
package hep.wired.heprep.representation;

import hep.graphics.heprep.HepRepInstance;
import hep.graphics.heprep.HepRepPoint;
import hep.wired.heprep.services.Attributes;
import hep.wired.heprep.services.GraphicsMode;
import hep.wired.heprep.services.Projection;
import hep.wired.heprep.util.PolyPoint;
import hep.wired.services.ViewPort;

import java.util.Iterator;

import org.freehep.graphics2d.VectorGraphics;

/**
 * Draws one or more cylinders, where a couple of HepRepPoints specify the faces
 * centers, with given "Radius", positioned using "Phi" and "Theta".
 * 
 * @author Mark Donszelmann
 * @version $Id: DrawAsCylinder.java 8598 2006-08-14 20:38:02Z duns $
 */

public class DrawAsCylinder extends AbstractDrawAs {

    private static final int MAX_FACES = 64;

    private double circle1[][] = new double[3][PolyPoint.ELLIPSEPOINTS];

    private double circle2[][] = new double[3][PolyPoint.ELLIPSEPOINTS];

    private double side[][][] = new double[MAX_FACES][3][2];

    public String getKey() {
        return "cylinder";
    }

    public String getName() {
        return "Cylinder";
    }

    public String getDescription() {
        return "Draws a HepRepInstance as a Cylinder.";
    }

    private void createCylinders(double x1, double y1, double z1, double r1,
            double phi1, double theta1, double x2, double y2, double z2,
            double r2, double phi2, double theta2, int faces) {
        double dx = x2 - x1;
        double dy = y2 - y1;
        double dz = z2 - z1;
        double dt = Math.sqrt(dx * dx + dz * dz);
        double phi = Math.atan2(dx, dz);
        double theta = -Math.atan2(dy, dt);

        PolyPoint.circle(r1, x1, y1, z1, phi + phi1, theta + theta1, circle1);
        PolyPoint.circle(r2, x2, y2, z2, phi + phi2, theta + theta2, circle2);
        PolyPoint.linesOnArcs(r1, r1, x1, y1, z1, phi + phi1, theta + theta1,
                0, r2, r2, x2, y2, z2, phi + phi2, theta + theta2, 0, 0,
                Math.PI * 2, faces, side);
    }

    public void draw(VectorGraphics graphics, HepRepInstance instance,
            Attributes atts, GraphicsMode mode, Projection projection,
            ViewPort viewPort) {

        Iterator iterator = instance.getPoints().iterator();
        if (!iterator.hasNext())
            return;
        HepRepPoint p = (HepRepPoint) iterator.next();
        int faces = Math.min(p.getAttValue("faces").getInteger(), MAX_FACES);
        double x0 = p.getX();
        double y0 = p.getY();
        double z0 = p.getZ();
        double r0 = p.getAttValue("radius").getDouble();
        double phi0 = p.getAttValue("phi").getDouble();
        double theta0 = p.getAttValue("theta").getDouble();
        while (iterator.hasNext()) {
            p = (HepRepPoint) iterator.next();
            double x = p.getX();
            double y = p.getY();
            double z = p.getZ();
            double r = p.getAttValue("radius").getDouble();
            double phi = p.getAttValue("phi").getDouble();
            double theta = p.getAttValue("theta").getDouble();

            createCylinders(x0, y0, z0, r0, phi0, theta0, x, y, z, r, phi,
                    theta, faces);

            circle1 = projection.transform(viewPort, circle1,
                    PolyPoint.ELLIPSEPOINTS);
            circle2 = projection.transform(viewPort, circle2,
                    PolyPoint.ELLIPSEPOINTS);

            if (atts.isFilled()) {
                graphics.setColor(atts.getFillColor());
                graphics.fillPolygon(circle1[U], circle1[V],
                        PolyPoint.ELLIPSEPOINTS);
                graphics.fillPolygon(circle2[U], circle2[V],
                        PolyPoint.ELLIPSEPOINTS);
                graphics.setColor(atts.getColor());
            }
            graphics.drawPolygon(circle1[U], circle1[V],
                    PolyPoint.ELLIPSEPOINTS);
            graphics.drawPolygon(circle2[U], circle2[V],
                    PolyPoint.ELLIPSEPOINTS);
            for (int i = 0; i < faces; i++) {
                side[i] = projection.transform(viewPort, side[i], 2);
                graphics.drawPolyline(side[i][U], side[i][V], 2);
            }

            x0 = x;
            y0 = y;
            z0 = z;
            r0 = r;
            phi0 = phi;
            theta0 = theta;
        }
    }

    public void frame(VectorGraphics graphics, HepRepInstance instance,
            Attributes atts, GraphicsMode mode, Projection projection,
            ViewPort viewPort) {

        graphics.setColor(atts.getFrameColor());
        graphics.setLineWidth(atts.getFrameWidth() * 2 + atts.getLineWidth());

        Iterator iterator = instance.getPoints().iterator();
        if (!iterator.hasNext())
            return;
        HepRepPoint p = (HepRepPoint) iterator.next();
        int faces = Math.min(p.getAttValue("faces").getInteger(), MAX_FACES);
        double x0 = p.getX();
        double y0 = p.getY();
        double z0 = p.getZ();
        double r0 = p.getAttValue("radius").getDouble();
        double phi0 = p.getAttValue("phi").getDouble();
        double theta0 = p.getAttValue("theta").getDouble();
        while (iterator.hasNext()) {
            p = (HepRepPoint) iterator.next();
            double x = p.getX();
            double y = p.getY();
            double z = p.getZ();
            double r = p.getAttValue("radius").getDouble();
            double phi = p.getAttValue("phi").getDouble();
            double theta = p.getAttValue("theta").getDouble();

            createCylinders(x0, y0, z0, r0, phi0, theta0, x, y, z, r, phi,
                    theta, faces);

            circle1 = projection.transform(viewPort, circle1,
                    PolyPoint.ELLIPSEPOINTS);
            circle1 = viewPort.transform(circle1, PolyPoint.ELLIPSEPOINTS);
            circle2 = projection.transform(viewPort, circle2,
                    PolyPoint.ELLIPSEPOINTS);
            circle2 = viewPort.transform(circle2, PolyPoint.ELLIPSEPOINTS);

            graphics.drawPolygon(circle1[U], circle1[V],
                    PolyPoint.ELLIPSEPOINTS);
            graphics.drawPolygon(circle2[U], circle2[V],
                    PolyPoint.ELLIPSEPOINTS);
            for (int i = 0; i < faces; i++) {
                side[i] = projection.transform(viewPort, side[i], 2);
                graphics.drawPolyline(side[i][U], side[i][V], 2);
            }

            x0 = x;
            y0 = y;
            z0 = z;
            r0 = r;
            phi0 = phi;
            theta0 = theta;
        }

        graphics.setLineWidth(atts.getLineWidth());
        graphics.setColor(atts.getColor());
    }
}
