// Copyright 2004-2005, FreeHEP.
package hep.wired.heprep.representation;

import hep.graphics.heprep.HepRepInstance;
import hep.graphics.heprep.HepRepPoint;
import hep.wired.heprep.services.Attributes;
import hep.wired.heprep.services.GraphicsMode;
import hep.wired.heprep.services.Projection;
import hep.wired.heprep.util.PolyPoint;
import hep.wired.services.ViewPort;

import java.util.Iterator;

import org.freehep.graphics2d.VectorGraphics;

/**
 * Draws one or more cylinders with elliptical faces, centered at HepRepPoints,
 * with given "Radius" and "Radius2", positioned using "Phi", "Theta" and
 * "Omega".
 * 
 * @author Mark Donszelmann
 * @version $Id: DrawAsEllipsePrism.java 8598 2006-08-14 20:38:02Z duns $
 */

public class DrawAsEllipsePrism extends AbstractDrawAs {

    private static final int MAX_FACES = 64;

    private double ellipse1[][] = new double[3][PolyPoint.ELLIPSEPOINTS];

    private double ellipse2[][] = new double[3][PolyPoint.ELLIPSEPOINTS];

    private double side[][][] = new double[MAX_FACES][3][2];

    public String getKey() {
        return "ellipseprism";
    }

    public String getName() {
        return "EllipsePrism";
    }

    public String getDescription() {
        return "Draws a HepRepInstance as a Cylinder with Elliptical ends.";
    }

    private void createEllipsePrism(double x1, double y1, double z1,
            double r11, double r12, double phi1, double theta1, double omega1,
            double x2, double y2, double z2, double r21, double r22,
            double phi2, double theta2, double omega2, int faces) {
        double dx = x2 - x1;
        double dy = y2 - y1;
        double dz = z2 - z1;
        double dt = Math.sqrt(dx * dx + dz * dz);
        double phi = Math.atan2(dx, dz);
        double theta = -Math.atan2(dy, dt);

        PolyPoint.ellipse(r11, r12, x1, y1, z1, phi + phi1, theta + theta1,
                omega1, ellipse1);
        PolyPoint.ellipse(r21, r22, x2, y2, z2, phi + phi2, theta + theta2,
                omega2, ellipse2);
        PolyPoint.linesOnArcs(r11, r12, x1, y1, z1, phi + phi1, theta + theta1,
                omega1, r21, r22, x2, y2, z2, phi + phi2, theta + theta2,
                omega2, 0, Math.PI * 2, faces, side);
    }

    public void draw(VectorGraphics graphics, HepRepInstance instance,
            Attributes atts, GraphicsMode mode, Projection projection,
            ViewPort viewPort) {

        Iterator iterator = instance.getPoints().iterator();
        if (!iterator.hasNext())
            return;
        HepRepPoint p = (HepRepPoint) iterator.next();
        int faces = Math.min(p.getAttValue("faces").getInteger(), MAX_FACES);
        double x0 = p.getX();
        double y0 = p.getY();
        double z0 = p.getZ();
        double r01 = p.getAttValue("radius").getDouble();
        double r02 = p.getAttValue("radius2").getDouble();
        double phi0 = p.getAttValue("phi").getDouble();
        double theta0 = p.getAttValue("theta").getDouble();
        double omega0 = p.getAttValue("omega").getDouble();
        while (iterator.hasNext()) {
            p = (HepRepPoint) iterator.next();
            double x = p.getX();
            double y = p.getY();
            double z = p.getZ();
            double r1 = p.getAttValue("radius").getDouble();
            double r2 = p.getAttValue("radius2").getDouble();
            double phi = p.getAttValue("phi").getDouble();
            double theta = p.getAttValue("theta").getDouble();
            double omega = p.getAttValue("omega").getDouble();

            createEllipsePrism(x0, y0, z0, r01, r02, phi0, theta0, omega0, x,
                    y, z, r1, r2, phi, theta, omega, faces);

            ellipse1 = projection.transform(viewPort, ellipse1,
                    PolyPoint.ELLIPSEPOINTS);
            ellipse2 = projection.transform(viewPort, ellipse2,
                    PolyPoint.ELLIPSEPOINTS);

            if (atts.isFilled()) {
                graphics.setColor(atts.getFillColor());
                graphics.fillPolygon(ellipse1[U], ellipse1[V],
                        PolyPoint.ELLIPSEPOINTS);
                graphics.fillPolygon(ellipse2[U], ellipse2[V],
                        PolyPoint.ELLIPSEPOINTS);
                graphics.setColor(atts.getColor());
            }
            graphics.drawPolygon(ellipse1[U], ellipse1[V],
                    PolyPoint.ELLIPSEPOINTS);
            graphics.drawPolygon(ellipse2[U], ellipse2[V],
                    PolyPoint.ELLIPSEPOINTS);
            for (int i = 0; i < faces; i++) {
                side[i] = projection.transform(viewPort, side[i], 2);
                graphics.drawPolyline(side[i][U], side[i][V], 2);
            }

            x0 = x;
            y0 = y;
            z0 = z;
            r01 = r1;
            r02 = r2;
            phi0 = phi;
            theta0 = theta;
            omega0 = omega;
        }
    }

    public void frame(VectorGraphics graphics, HepRepInstance instance,
            Attributes atts, GraphicsMode mode, Projection projection,
            ViewPort viewPort) {

        graphics.setColor(atts.getFrameColor());
        graphics.setLineWidth(atts.getFrameWidth() * 2 + atts.getLineWidth());

        Iterator iterator = instance.getPoints().iterator();
        if (!iterator.hasNext())
            return;
        HepRepPoint p = (HepRepPoint) iterator.next();
        int faces = Math.min(p.getAttValue("faces").getInteger(), MAX_FACES);
        double x0 = p.getX();
        double y0 = p.getY();
        double z0 = p.getZ();
        double r01 = p.getAttValue("radius").getDouble();
        double r02 = p.getAttValue("radius2").getDouble();
        double phi0 = p.getAttValue("phi").getDouble();
        double theta0 = p.getAttValue("theta").getDouble();
        double omega0 = p.getAttValue("omega").getDouble();
        while (iterator.hasNext()) {
            p = (HepRepPoint) iterator.next();
            double x = p.getX();
            double y = p.getY();
            double z = p.getZ();
            double r1 = p.getAttValue("radius").getDouble();
            double r2 = p.getAttValue("radius2").getDouble();
            double phi = p.getAttValue("phi").getDouble();
            double theta = p.getAttValue("theta").getDouble();
            double omega = p.getAttValue("omega").getDouble();

            createEllipsePrism(x0, y0, z0, r01, r02, phi0, theta0, omega0, x,
                    y, z, r1, r2, phi, theta, omega, faces);

            ellipse1 = projection.transform(viewPort, ellipse1,
                    PolyPoint.ELLIPSEPOINTS);
            ellipse2 = projection.transform(viewPort, ellipse2,
                    PolyPoint.ELLIPSEPOINTS);

            graphics.drawPolygon(ellipse1[U], ellipse1[V],
                    PolyPoint.ELLIPSEPOINTS);
            graphics.drawPolygon(ellipse2[U], ellipse2[V],
                    PolyPoint.ELLIPSEPOINTS);
            for (int i = 0; i < faces; i++) {
                side[i] = projection.transform(viewPort, side[i], 2);
                graphics.drawPolyline(side[i][U], side[i][V], 2);
            }

            x0 = x;
            y0 = y;
            z0 = z;
            r01 = r1;
            r02 = r2;
            phi0 = phi;
            theta0 = theta;
            omega0 = omega;
        }

        graphics.setLineWidth(atts.getLineWidth());
        graphics.setColor(atts.getColor());
    }
}
