package hep.wired.heprep.tree;

import java.util.*;

import hep.graphics.heprep.HepRep;
import hep.graphics.heprep.HepRepInstance;
import hep.graphics.heprep.HepRepInstanceTree;
import hep.graphics.heprep.HepRepType;
import hep.graphics.heprep.HepRepTypeTree;
import hep.wired.heprep.util.WiredHepRepUtil;
import hep.wired.util.tree.WiredTreeModel;
import hep.wired.util.tree.WiredTreeNode;
import hep.wired.util.tree.WiredTreeNode.Flag;
import javax.swing.tree.TreePath;
import org.freehep.application.Application;
import org.freehep.application.PropertyUtilities;

/**
 * HepRep-specific extension of {@link WiredTreeModel}.
 *
 * @author Dmitry Onoprienko
 * @version $Id: HepRepTreeModel.java 13864 2011-08-12 18:38:41Z onoprien $
 */
public class HepRepTreeModel extends WiredTreeModel {

// -- Private parts : ----------------------------------------------------------

  protected boolean _isTypeTree;

// -- Construction : -----------------------------------------------------------

  public HepRepTreeModel(HepRep heprep, boolean isTypeTree) {
     super(null);
    _isTypeTree = isTypeTree;
    if (heprep != null) setRoot(createNode(heprep));
  }

  public HepRepTreeModel(HepRepTreeModel model) {
     super(null);
    _isTypeTree = model._isTypeTree;
    Object rt = model.getRoot().getPayload();
    if (rt != null) setRoot(createNode(rt));
    copySettings(model.getRoot(), getRoot());
  }

  private WiredTreeNode createNode(Object payload) {
    WiredTreeNode node = new WiredTreeNode(payload);
    ArrayList<Object> children = new ArrayList(getChildren(payload));
    Collections.sort(children, new Comparator<Object>() {
      public int compare(Object o1, Object o2) {
        return getName(o1).compareTo(getName(o2));
      }
    });
    for (Object child : children) {
      WiredTreeNode childNode = createNode(child);
      node.add(childNode);
    }
    return node;
  }

  private void copySettings(WiredTreeNode fromNode, WiredTreeNode toNode) {
    toNode.setFlags(fromNode.getFlags());
    int n = toNode.getChildCount();
    for (int i=0; i<n; i++) {
      copySettings((WiredTreeNode)fromNode.getChildAt(i), (WiredTreeNode)toNode.getChildAt(i));
    }
  }

// -- Getters : ----------------------------------------------------------------

  public int getMaximumDepth() {
    return root == null ? 0 : Math.max(0, getRoot().getDepth());
  }

  /**
   * Returns a set of objects in "marked" nodes, excluding levels 0 and 1.
   */
  public Set getMarkedObjects() {
    return getMarkedObjects(-1);
  }

  /**
   * Returns a set of objects in "marked" nodes in levels 2 through <tt>maxLevel</tt>.
   * Layers are numbered starting with 0.
   */
  public Set getMarkedObjects(int maxLevel) {
//    final Set out = new HashSet();
    final Set out = Collections.newSetFromMap(new IdentityHashMap());
//    final Set out = _isTypeTree ? new HashSet() : Collections.newSetFromMap(new IdentityHashMap());
    TreeVisitor v = new TreeVisitor() {
      public boolean visit(WiredTreeNode node) {
        if (node.isMarkedNode() && ! node.isHidden() && node.getLevel()>1) out.add(node.getPayload());
        return node.isMarkedBranch();
      }
    };
    v.visitNodes(this, 1, maxLevel);
    return out;
  }

  /**
   * Returns a list of expanded paths, in order of decreasing path length.
   * Path is considered expanded if its last component is expanded - this is different
   * from JTree <tt>isExpanded()</tt>, where all path components are required to be expanded.
   */
  public List<TreePath> getExpandedPaths() {
    final ArrayList<TreePath> out = new ArrayList();
    TreeVisitor v = new TreeVisitor() {
      public boolean visit(TreePath path) {
        WiredTreeNode node = (WiredTreeNode) path.getLastPathComponent();
        if (node.isExpanded()) out.add(path);
        return true;
      }
    };
    v.visitNodes(this);
    Collections.sort(out, new Comparator<TreePath>() {
      public int compare(TreePath path1, TreePath path2) {
        return path1.getPathCount() - path2.getPathCount();
      }
    });
    return out;
  }

// -- Utility methods : --------------------------------------------------------

  /**
   * Returns the name of the specified HepRep object - to be used in display and sorting.
   */
  static public String getName(Object payload) {
    if (payload instanceof HepRepInstance) {
      HepRepInstance instance = (HepRepInstance) payload;
      return instance.getType().getName();
    } else if (payload instanceof HepRepType) {
      HepRepType type = (HepRepType) payload;
      return type.getName();
    } else if (payload instanceof HepRepInstanceTree) {
      HepRepInstanceTree tree = (HepRepInstanceTree) payload;
      return tree.getName();
    } else if (payload instanceof HepRepTypeTree) {
      HepRepTypeTree tree = (HepRepTypeTree) payload;
      return tree.getName();
    } else if (payload instanceof HepRep) {
      return "HepRep";
    }
    return payload != null ? payload.toString() : "null";
  }

//  public TreePath getPath(Object payload) {
//
//  }

// -- Storing/restoring tree expansion and mark state to jas3 properties : -----

  synchronized public void saveTreeState(String key) {

    final HashSet<String> unmarkedNodes = new HashSet();
    final HashSet<String> unmarkedBranches = new HashSet();
    final HashSet<String> expanded = new HashSet();

    TreeVisitor v = new TreeVisitor() {
      public boolean visit(TreePath path) {
        WiredTreeNode node = (WiredTreeNode) path.getLastPathComponent();
        boolean un = !node.isMarkedNode();
        boolean ub = !node.isLeaf() && !node.isMarkedBranch();
        boolean ex = node.isExpanded();
        if (ex || un || ub) {
          String spath = pathToString(path);
          if (un) unmarkedNodes.add(spath);
          if (ub) unmarkedBranches.add(spath);
          if (ex) expanded.add(spath);
        }
        return true;
      }
    };
    v.visitPaths(this, 1, -1);

    Properties user = Application.getApplication().getUserProperties();
    PropertyUtilities.setStringCollection(user, key + ".un", unmarkedNodes);
    PropertyUtilities.setStringCollection(user, key + ".ub", unmarkedBranches);
    PropertyUtilities.setStringCollection(user, key + ".ex", expanded);
  }

  synchronized public void restoreTreeState(String key) {

    Properties user = Application.getApplication().getUserProperties();
    final HashSet<String>  unmarkedNodes = new HashSet(PropertyUtilities.getStringCollection(user, key + ".un", Collections.EMPTY_SET));
    final HashSet<String>  unmarkedBranches = new HashSet(PropertyUtilities.getStringCollection(user, key + ".ub", Collections.EMPTY_SET));
    final HashSet<String>  expanded = new HashSet(PropertyUtilities.getStringCollection(user, key + ".ex", Collections.EMPTY_SET));

    TreeVisitor v = new TreeVisitor() {
      public boolean visit(TreePath path) {
        WiredTreeNode node = (WiredTreeNode) path.getLastPathComponent();
        String spath = pathToString(path);
        node.setFlag(Flag.MARKED_NODE, !unmarkedNodes.contains(spath));
        node.setFlag(Flag.MARKED_BRANCH, !unmarkedBranches.contains(spath));
        node.setFlag(Flag.EXPANDED, expanded.contains(spath));
        node.removeFlag(Flag.HAS_UNMARKED_DESCENDANTS_VALID);
        return true;
      }
    };
    v.visitPaths(this, 1, -1);
  }

  private String pathToString(TreePath path) {
    Object[] pp = path.getPath();
    String[] elements = new String[pp.length];
    for (int i=0; i<pp.length; i++) elements[i] = WiredHepRepUtil.getHepRepName(((WiredTreeNode)pp[i]).getPayload());
    StringBuilder s = new StringBuilder();
    for (int i = 0; i < elements.length; i++) {
      String name = elements[i];
      name = name.replaceAll("\\\\", "\\\\");  // escape \
      name = name.replaceAll("/", "\\/");  // escape /
      // escape =
      name = name.replaceAll("=", "\\=");

      s.append("/");
      s.append(name);
    }
    return s.toString();
  }

  /**
   * Creates and returns a new tree model containing objects from the specified <tt>heprep</tt>,
   * with expansion and mark states of all nodes copied from this <tt>HepRepTreeModel</tt>.
   * If the structure of the supplied <tt>heprep</tt> object is different from the one used
   * to create this tree, returns <tt>null</tt>.
   */
  public HepRepTreeModel clone(HepRep heprep) {

    final LinkedHashMap<String, EnumSet<Flag>> path2flags = new LinkedHashMap();

    TreeVisitor v = new TreeVisitor() {
      public boolean visit(TreePath path) {
        WiredTreeNode node = (WiredTreeNode) path.getLastPathComponent();
        String spath = pathToString(path);
        path2flags.put(spath, node.getFlags());
        return true;
      }
    };
    v.visitPaths(this);

    HepRepTreeModel out = new HepRepTreeModel(heprep, _isTypeTree);

    v = new TreeVisitor() {
      public boolean visit(TreePath path) {
        WiredTreeNode node = (WiredTreeNode) path.getLastPathComponent();
        String spath = pathToString(path);
        EnumSet<Flag> flags = path2flags.get(spath);
        if (flags == null) {
          throw new IllegalArgumentException();
        } else {
          node.setFlags(flags);
        }
        return true;
      }
    };
    try {
      v.visitPaths(out);
    } catch (IllegalArgumentException x) {
      return null;
    }

    return out;
  }

// -- Local methods : ----------------------------------------------------------

  private List<Object> getChildren(Object payload) {
    List children;
    if (payload instanceof HepRepInstance) {
      children = ((HepRepInstance)payload).getInstances();
    } else if (payload instanceof HepRepType) {
      children = ((HepRepType)payload).getTypeList();
    } else if (payload instanceof HepRepTypeTree) {
      children = ((HepRepTypeTree)payload).getTypeList();
    } else if (payload instanceof HepRepInstanceTree) {
      children = ((HepRepInstanceTree)payload).getInstances();
    } else if(payload instanceof HepRep) {
      children = _isTypeTree ? ((HepRep)payload).getTypeTreeList() : ((HepRep)payload).getInstanceTreeList() ;
    } else {
      throw new RuntimeException("Cannot add objects of type "+ payload.getClass().getName() +" to WiredHepRepTree.");
    }
    return (List<Object>) children;
  }

// -----------------------------------------------------------------------------
}
