// Copyright 2004, FreeHEP.
package hep.wired.heprep.util;

import java.awt.geom.CubicCurve2D;
import java.awt.geom.Point2D;
import java.io.IOException;

import org.freehep.graphicsio.QuadToCubicPathConstructor;

/**
 * Constructs a path while keeping track of the nearest point on that path to an
 * arbitrary given point.
 * 
 * @author Mark Donszelmann
 * @version $Id: NearestPointPathConstructor.java 8598 2006-08-14 20:38:02Z duns $
 */
public class NearestPointPathConstructor extends QuadToCubicPathConstructor {

    private Point2D arbitraryPoint;

    private Point2D nearestPoint = new Point2D.Double();

    private Point2D nearestPointOnPath = new Point2D.Double();

    private double distanceSq = Double.MAX_VALUE;

    private transient Point2D p1 = new Point2D.Double();

    private transient Point2D p2 = new Point2D.Double();

    private transient CubicCurve2D c = new CubicCurve2D.Double();

    /**
     * Creates a path constructor with given arbitrary point
     */
    public NearestPointPathConstructor(Point2D arbitraryPoint) {
        this.arbitraryPoint = arbitraryPoint;
    }

    /**
     * Adds a line to the path, called from addPath.
     */
    public void line(double x, double y) throws IOException {
        p1.setLocation(currentX, currentY);
        p2.setLocation(x, y);
        check(NearestPoint.onLine(p1, p2, arbitraryPoint, nearestPoint));
        super.line(x, y);
    }

    /**
     * Adds a cubic bezier to the path, called from addPath.
     */
    public void cubic(double x1, double y1, double x2, double y2, double x3,
            double y3) throws IOException {
        c.setCurve(currentX, currentY, x1, y1, x2, y2, x3, y3);
        check(NearestPoint.onCurve(c, arbitraryPoint, nearestPoint));
        super.cubic(x1, y1, x2, y2, x3, y3);
    }

    /**
     * Closes the current path, called from addPath.
     */
    public void closePath(double x0, double y0) throws IOException {
        p1.setLocation(currentX, currentY);
        p2.setLocation(x0, y0);
        check(NearestPoint.onLine(p1, p2, arbitraryPoint, nearestPoint));
        super.closePath(x0, y0);
    }

    /**
     * Returns nearest point on path, nearest to the given arbitrary point, or
     * null if path is empty.
     */
    public Point2D getNearestPoint() {
        return distanceSq == Double.MAX_VALUE ? null : nearestPointOnPath;
    }

    /**
     * Returns squared distance from the arbitrary point to the nearest point on
     * path. If path is empty, -1 is returned.
     */
    public double distanceSq() {
        return distanceSq == Double.MAX_VALUE ? -1 : distanceSq;
    }

    private void check(double dsq) {
        if (dsq < distanceSq) {
            distanceSq = dsq;
            nearestPointOnPath.setLocation(nearestPoint);
            pointFound(new Point2D.Double(nearestPoint.getX(), nearestPoint
                    .getY()), distanceSq, true);
        } else if (dsq == distanceSq) {
            pointFound(new Point2D.Double(nearestPoint.getX(), nearestPoint
                    .getY()), distanceSq, false);
        }
    }

    /**
     * Called when a closer or equal point is found, so that the subclass can
     * use this to track some other information, such as keeping a reference to
     * the closest object(s).
     */
    protected void pointFound(Point2D nearestPoint, double distanceSq,
            boolean closer) {
    }
}
